import logging
import os
import pathlib

import keras
from keras import Model
from keras.layers import LeakyReLU

from exoml.ml import thread_config
from exoml.detrend.detrend_generator import DetrendModelGenerator
from exoml.ml.loss.unsupervised import intransit_weighted_mean_squared_error
from exoml.ml.metrics.fixed_mean import FixedMean

from exoml.ml.model.base_model import BaseModel, HyperParams

thread_config.setup_threads(5)


class DETREND(BaseModel):
    def __init__(self, hyperparameters, name='DETREND', input_size=(20610, 8)):
        super().__init__(name, input_size, None, None, hyperparameters)

    def build(self):
        # (time, flux, flux_err, centroidx, centroidy, motionx, motiony, bck)
        autoencoder_layer1_strides = 10
        autoencoder_layer1_filters = 5000
        autoencoder_layer1_ks = 100
        autoencoder_layer2_strides = 5
        autoencoder_layer2_filters = 1250
        autoencoder_layer2_ks = 33
        autoencoder_layer3_strides = 5
        autoencoder_layer3_filters = 420
        autoencoder_layer3_ks = 15
        autoencoder_layer4_strides = 2
        autoencoder_layer4_filters = 128
        autoencoder_layer4_ks = 9
        autoencoder_layer5_strides = 2
        autoencoder_layer5_filters = 64
        autoencoder_layer5_ks = 7
        autoencoder_layer6_strides = 1
        autoencoder_layer6_filters = 32
        autoencoder_layer6_ks = 5
        input = keras.Input(shape=(self.input_size))
        self.batch_norm_input = keras.layers.normalization.batch_normalization.BatchNormalization()(input)
        self.enc_layer1 = keras.layers.SpatialDropout1D(rate=0.1)(input)
        self.enc_layer1_r = keras.layers.Conv1D(filters=autoencoder_layer1_filters, kernel_size=autoencoder_layer1_ks,
                                                padding="same", activation=LeakyReLU(0.01))(self.enc_layer1)
        self.enc_layer1 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.enc_layer1_r)
        self.enc_layer1 = keras.layers.MaxPooling1D(pool_size=50, strides=autoencoder_layer1_strides, padding="same")(self.enc_layer1)
        self.enc_layer1 = keras.layers.Dropout(rate=0.1)(self.enc_layer1)
        self.enc_layer2_r = keras.layers.Conv1D(filters=autoencoder_layer2_filters, kernel_size=autoencoder_layer2_ks,
                                                padding="same", activation=LeakyReLU(0.01))(self.enc_layer1)
        self.enc_layer2 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.enc_layer2_r)
        self.enc_layer2 = keras.layers.MaxPooling1D(pool_size=20, strides=autoencoder_layer2_strides, padding="same")(self.enc_layer2)
        self.enc_layer2 = keras.layers.Dropout(rate=0.1)(self.enc_layer2)
        self.enc_layer3_r = keras.layers.Conv1D(filters=autoencoder_layer3_filters, kernel_size=autoencoder_layer3_ks,
                                                padding="same", activation=LeakyReLU(0.01))(self.enc_layer2)
        self.enc_layer3 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.enc_layer3_r)
        self.enc_layer3 = keras.layers.MaxPooling1D(pool_size=15, strides=autoencoder_layer3_strides, padding="same")(self.enc_layer3)
        self.enc_layer3 = keras.layers.Dropout(rate=0.1)(self.enc_layer3)
        self.enc_layer4_r = keras.layers.Conv1D(filters=autoencoder_layer4_filters, kernel_size=autoencoder_layer4_ks,
                                                padding="same", activation=LeakyReLU(0.01))(self.enc_layer3)
        self.enc_layer4 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.enc_layer4_r)
        self.enc_layer4 = keras.layers.MaxPooling1D(pool_size=10, strides=autoencoder_layer4_strides, padding="same")(self.enc_layer4)
        self.enc_layer4 = keras.layers.Dropout(rate=0.1)(self.enc_layer4)
        self.enc_layer5_r = keras.layers.Conv1D(filters=autoencoder_layer5_filters, kernel_size=autoencoder_layer5_ks,
                                                padding="same", activation=LeakyReLU(0.01))(self.enc_layer4)
        self.enc_layer5 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.enc_layer5_r)
        self.enc_layer5 = keras.layers.MaxPooling1D(pool_size=4, strides=autoencoder_layer5_strides, padding="same")(self.enc_layer5)
        self.enc_layer5 = keras.layers.Dropout(rate=0.1)(self.enc_layer5)
        self.enc_layer6_r = keras.layers.Conv1D(filters=autoencoder_layer6_filters, kernel_size=autoencoder_layer6_ks,
                                                padding="same", activation=LeakyReLU(0.01))(self.enc_layer5)
        self.enc_layer6 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.enc_layer6_r)
        self.enc_layer6 = keras.layers.MaxPooling1D(pool_size=3, strides=autoencoder_layer6_strides, padding="same")(self.enc_layer6)
        self.enc_layer6 = keras.layers.Dropout(rate=0.1)(self.enc_layer6)
        self.dec_layer6 = keras.layers.UpSampling1D(autoencoder_layer6_strides)(self.enc_layer6)
        self.dec_layer6 = keras.layers.Conv1DTranspose(filters=autoencoder_layer6_filters, kernel_size=autoencoder_layer6_ks,
                                                       padding="same", activation=LeakyReLU(0.01))(self.dec_layer6)
        self.dec_layer6 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.dec_layer6)
        self.dec_layer6 = keras.layers.Add()([self.enc_layer6_r, self.dec_layer6])
        self.dec_layer5 = keras.layers.UpSampling1D(autoencoder_layer5_strides)(self.dec_layer6)
        self.dec_layer5 = keras.layers.Conv1DTranspose(filters=autoencoder_layer5_filters,
                                                       kernel_size=autoencoder_layer5_ks, padding="same",
                                                       activation=LeakyReLU(0.01))(self.dec_layer5)
        self.dec_layer5 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.dec_layer5)
        self.dec_layer5 = keras.layers.Add()([self.enc_layer5_r, self.dec_layer5])
        self.dec_layer4 = keras.layers.UpSampling1D(autoencoder_layer4_strides)(self.dec_layer5)
        self.dec_layer4 = keras.layers.Cropping1D(cropping=(0, 1))(self.dec_layer4)
        self.dec_layer4 = keras.layers.Conv1DTranspose(filters=autoencoder_layer4_filters,
                                                       kernel_size=autoencoder_layer4_ks, padding="same",
                                                       activation=LeakyReLU(0.01))(self.dec_layer4)
        self.dec_layer4 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.dec_layer4)
        self.dec_layer4 = keras.layers.Add()([self.enc_layer4_r, self.dec_layer4])
        self.dec_layer3 = keras.layers.UpSampling1D(autoencoder_layer3_strides)(self.dec_layer4)
        self.dec_layer3 = keras.layers.Cropping1D(cropping=1)(self.dec_layer3)
        self.dec_layer3 = keras.layers.Conv1DTranspose(filters=autoencoder_layer3_filters,
                                                       kernel_size=autoencoder_layer3_ks, padding="same",
                                                       activation=LeakyReLU(0.01))(self.dec_layer3)
        self.dec_layer3 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.dec_layer3)
        self.dec_layer3 = keras.layers.Add()([self.enc_layer3_r, self.dec_layer3])
        self.dec_layer2 = keras.layers.UpSampling1D(autoencoder_layer2_strides)(self.dec_layer3)
        self.dec_layer2 = keras.layers.Cropping1D(cropping=2)(self.dec_layer2)
        self.dec_layer2 = keras.layers.Conv1DTranspose(filters=autoencoder_layer2_filters,
                                                       kernel_size=autoencoder_layer2_ks, padding="same",
                                                       activation=LeakyReLU(0.01))(self.dec_layer2)
        self.dec_layer2 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.dec_layer2)
        self.dec_layer2 = keras.layers.Add()([self.enc_layer2_r, self.dec_layer2])
        self.dec_layer1 = keras.layers.UpSampling1D(autoencoder_layer1_strides)(self.dec_layer2)
        self.dec_layer1 = keras.layers.Conv1DTranspose(filters=autoencoder_layer1_filters,
                                                       kernel_size=autoencoder_layer1_ks, padding="same",
                                                       activation=LeakyReLU(0.01))(self.dec_layer1)
        self.dec_layer1 = keras.layers.normalization.batch_normalization.BatchNormalization()(self.dec_layer1)
        self.dec_layer1 = keras.layers.Add()([self.enc_layer1_r, self.dec_layer1])
        self.linear_proj = keras.layers.GlobalAveragePooling1D()(self.dec_layer1)
        self.linear_proj = keras.layers.Dense(self.input_size[0], activation=LeakyReLU(0.01))(self.linear_proj)
        self.linear_proj = keras.layers.Dropout(rate=0.1)(self.linear_proj)
        self.linear_proj = keras.layers.Dense(self.input_size[0], activation="linear")(self.linear_proj)
        self.set_model(Model(inputs=input, outputs=self.linear_proj))
        return self

    def load_training_set(self, **kwargs):
        training_dir = kwargs.get('training_dir')
        return [str(file) for file in list(pathlib.Path(training_dir).glob('*_lc.csv'))]

    def instance_generator(self, dataset, dir, batch_size, input_sizes, type_to_label, zero_epsilon):
        return DetrendModelGenerator(dataset, batch_size, input_sizes, zero_epsilon)

    def instance_loss_accuracy(self):
        #loss = tf.keras.losses.MeanAbsolutePercentageError(reduction=losses_utils.ReductionV2.AUTO)
        loss = intransit_weighted_mean_squared_error
        accuracy = FixedMean()
        return loss, accuracy

    def instance_metrics(self):
        return []

    def plot_metrics(self, model_dir, steps_per_epoch, with_validation=False):
        return []



import abc
import tensorflow as tf
import numpy as np
import tensorflow.keras.backend as K

from keras import backend
from keras.utils import metrics_utils
from keras.utils.generic_utils import to_list
from keras_core import Metric
from keras_core.src.utils import tf_utils
#from tf_keras.src.metrics import base_metric


#from tf_keras.src.metrics import base_metric
# from keras.src import backend
# from keras.src.utils import tf_utils
# from keras.metrics import Metric
# from keras.src.utils import metrics_utils
# from keras.src.utils.generic_utils import to_list
# from keras.src.metrics import base_metric
# from keras.src import backend

def inject_mesh(init_method):
    """Inject DTensor mesh information to an object.

    This is useful for keras object like `Metric` and `Optimizer` which need
    DTensor mesh to create the weights, but doesn't want to change the current
    public API interface.

    This is for temporary usage and eventually the mesh/layout information will
    be public arguments in the `__init__` method.

    Sample usage:
    ```python
    class Accuracy(tf.keras.metrics.Metric):

      @inject_mesh
      def __init__(self, name='accuracy', dtype=None):
         super().__init__(**kwargs)

      acc = Accuracy(mesh=mesh)
      assert acc._mesh == mesh
    ```

    Args:
      init_method: the `__init__` method of the Keras class to annotate.

    Returns:
      the annotated __init__ method.
    """

    def _wrap_function(instance, *args, **kwargs):
        mesh = kwargs.pop("mesh", None)
        # Note that the injection of _mesh need to happen before the invocation
        # of __init__, since the class might need the mesh to create weights in
        # the __init__.
        if mesh is not None:
            instance._mesh = mesh
        init_method(instance, *args, **kwargs)

    return tf.__internal__.decorator.make_decorator(
        target=init_method, decorator_func=_wrap_function
    )


class SensitivitySpecificityBase(Metric, metaclass=abc.ABCMeta):
    """Abstract base class for computing sensitivity and specificity.

    For additional information about specificity and sensitivity, see
    [the following](https://en.wikipedia.org/wiki/Sensitivity_and_specificity).
    """

    def __init__(
        self, value, num_thresholds=200, class_id=None, name=None, dtype=None
    ):
        super().__init__(name=name, dtype=dtype)
        if num_thresholds <= 0:
            raise ValueError(
                "Argument `num_thresholds` must be an integer > 0. "
                f"Received: num_thresholds={num_thresholds}"
            )
        self.value = value
        self.class_id = class_id
        self.true_positives = self.add_weight(
            name="true_positives", shape=(num_thresholds,), initializer="zeros"
        )
        self.true_negatives = self.add_weight(
            name="true_negatives", shape=(num_thresholds,), initializer="zeros"
        )
        self.false_positives = self.add_weight(
            name="false_positives", shape=(num_thresholds,), initializer="zeros"
        )
        self.false_negatives = self.add_weight(
            name="false_negatives", shape=(num_thresholds,), initializer="zeros"
        )

        # Compute `num_thresholds` thresholds in [0, 1]
        if num_thresholds == 1:
            self.thresholds = [0.5]
            self._thresholds_distributed_evenly = False
        else:
            thresholds = [
                (i + 1) * 1.0 / (num_thresholds - 1)
                for i in range(num_thresholds - 2)
            ]
            self.thresholds = [0.0] + thresholds + [1.0]
            self._thresholds_distributed_evenly = True

    def update_state(self, y_true, y_pred, sample_weight=None):
        """Accumulates confusion matrix statistics.

        Args:
          y_true: The ground truth values.
          y_pred: The predicted values.
          sample_weight: Optional weighting of each example. Defaults to 1. Can
            be a `Tensor` whose rank is either 0, or the same rank as `y_true`,
            and must be broadcastable to `y_true`.

        Returns:
          Update op.
        """
        return metrics_utils.update_confusion_matrix_variables(
            {
                metrics_utils.ConfusionMatrix.TRUE_POSITIVES: self.true_positives,  # noqa: E501
                metrics_utils.ConfusionMatrix.TRUE_NEGATIVES: self.true_negatives,  # noqa: E501
                metrics_utils.ConfusionMatrix.FALSE_POSITIVES: self.false_positives,  # noqa: E501
                metrics_utils.ConfusionMatrix.FALSE_NEGATIVES: self.false_negatives,  # noqa: E501
            },
            y_true,
            y_pred,
            thresholds=self.thresholds,
            thresholds_distributed_evenly=self._thresholds_distributed_evenly,
            class_id=self.class_id,
            sample_weight=sample_weight,
        )

    def reset_state(self):
        num_thresholds = len(self.thresholds)
        confusion_matrix_variables = (
            self.true_positives,
            self.true_negatives,
            self.false_positives,
            self.false_negatives,
        )
        backend.batch_set_value(
            [
                (v, np.zeros((num_thresholds,)))
                for v in confusion_matrix_variables
            ]
        )

    def get_config(self):
        config = {"class_id": self.class_id}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def _find_max_under_constraint(self, constrained, dependent, predicate):
        """Returns the maximum of dependent_statistic that satisfies the
        constraint.

        Args:
          constrained: Over these values the constraint
            is specified. A rank-1 tensor.
          dependent: From these values the maximum that satiesfies the
            constraint is selected. Values in this tensor and in
            `constrained` are linked by having the same threshold at each
            position, hence this tensor must have the same shape.
          predicate: A binary boolean functor to be applied to arguments
          `constrained` and `self.value`, e.g. `tf.greater`.

        Returns:
          maximal dependent value, if no value satiesfies the constraint 0.0.
        """
        feasible = tf.where(predicate(constrained, self.value))
        feasible_exists = tf.greater(tf.size(feasible), 0)
        max_dependent = tf.reduce_max(tf.gather(dependent, feasible))

        return tf.where(feasible_exists, max_dependent, 0.0)



class ThresholdAtPrecision(SensitivitySpecificityBase):
    """Computes best recall where precision is >= specified value.

    For a given score-label-distribution the required precision might not
    be achievable, in this case 0.0 is returned as recall.

    This metric creates four local variables, `true_positives`,
    `true_negatives`, `false_positives` and `false_negatives` that are used to
    compute the recall at the given precision. The threshold for the given
    precision value is computed and used to evaluate the corresponding recall.

    If `sample_weight` is `None`, weights default to 1.
    Use `sample_weight` of 0 to mask values.

    If `class_id` is specified, we calculate precision by considering only the
    entries in the batch for which `class_id` is above the threshold
    predictions, and computing the fraction of them for which `class_id` is
    indeed a correct label.

    Args:
      precision: A scalar value in range `[0, 1]`.
      num_thresholds: (Optional) Defaults to 200. The number of thresholds to
        use for matching the given precision.
      class_id: (Optional) Integer class ID for which we want binary metrics.
        This must be in the half-open interval `[0, num_classes)`, where
        `num_classes` is the last dimension of predictions.
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.RecallAtPrecision(0.8)
    >>> m.update_state([0, 0, 1, 1], [0, 0.5, 0.3, 0.9])
    >>> m.result().numpy()
    0.5

    >>> m.reset_state()
    >>> m.update_state([0, 0, 1, 1], [0, 0.5, 0.3, 0.9],
    ...                sample_weight=[1, 0, 0, 1])
    >>> m.result().numpy()
    1.0

    Usage with `compile()` API:

    ```python
    model.compile(
        optimizer='sgd',
        loss='mse',
        metrics=[tf.keras.metrics.RecallAtPrecision(precision=0.8)])
    ```
    """

    @inject_mesh
    def __init__(
        self,
        precision,
        num_thresholds=200,
        class_id=None,
        name=None,
        dtype=None,
    ):
        if precision < 0 or precision > 1:
            raise ValueError(
                "Argument `precision` must be in the range [0, 1]. "
                f"Received: precision={precision}"
            )
        self.precision = precision
        self.num_thresholds = num_thresholds
        super().__init__(
            value=precision,
            num_thresholds=num_thresholds,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )

    def result(self):
        precisions = tf.math.divide_no_nan(
            self.true_positives,
            tf.math.add(self.true_positives, self.false_positives),
        )
        recalls = tf.math.divide_no_nan(
            self.true_positives,
            tf.math.add(self.true_positives, self.false_negatives),
        )
        return self._find_max_threshold_under_constraint(
            precisions, recalls, tf.greater_equal
        )

    def _find_max_threshold_under_constraint(self, constrained, dependent, predicate):
        """Returns the maximum of dependent_statistic that satisfies the
        constraint.

        Args:
          constrained: Over these values the constraint
            is specified. A rank-1 tensor.
          dependent: From these values the maximum that satiesfies the
            constraint is selected. Values in this tensor and in
            `constrained` are linked by having the same threshold at each
            position, hence this tensor must have the same shape.
          predicate: A binary boolean functor to be applied to arguments
          `constrained` and `self.value`, e.g. `tf.greater`.

        Returns:
          maximal dependent value, if no value satiesfies the constraint 0.0.
        """
        variable_dtype = self.true_positives.dtype
        thresholds = tf.convert_to_tensor(self.thresholds, dtype=variable_dtype)
        feasible = tf.where(predicate(constrained, self.value))
        feasible_exists = tf.greater(tf.size(feasible), 0)
        try:
            return tf.where(feasible_exists, tf.gather(thresholds, tf.gather(feasible, [0])), 0.0)
        except Exception as e:
            print("Error " + str(e))
            return 0.0 

    def get_config(self):
        config = {
            "num_thresholds": self.num_thresholds,
            "precision": self.precision,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class ThresholdAtRecall(SensitivitySpecificityBase):
    """Computes best recall where precision is >= specified value.

    For a given score-label-distribution the required precision might not
    be achievable, in this case 0.0 is returned as recall.

    This metric creates four local variables, `true_positives`,
    `true_negatives`, `false_positives` and `false_negatives` that are used to
    compute the recall at the given precision. The threshold for the given
    precision value is computed and used to evaluate the corresponding recall.

    If `sample_weight` is `None`, weights default to 1.
    Use `sample_weight` of 0 to mask values.

    If `class_id` is specified, we calculate precision by considering only the
    entries in the batch for which `class_id` is above the threshold
    predictions, and computing the fraction of them for which `class_id` is
    indeed a correct label.

    Args:
      recall: A scalar value in range `[0, 1]`.
      num_thresholds: (Optional) Defaults to 200. The number of thresholds to
        use for matching the given precision.
      class_id: (Optional) Integer class ID for which we want binary metrics.
        This must be in the half-open interval `[0, num_classes)`, where
        `num_classes` is the last dimension of predictions.
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.RecallAtPrecision(0.8)
    >>> m.update_state([0, 0, 1, 1], [0, 0.5, 0.3, 0.9])
    >>> m.result().numpy()
    0.5

    >>> m.reset_state()
    >>> m.update_state([0, 0, 1, 1], [0, 0.5, 0.3, 0.9],
    ...                sample_weight=[1, 0, 0, 1])
    >>> m.result().numpy()
    1.0

    Usage with `compile()` API:

    ```python
    model.compile(
        optimizer='sgd',
        loss='mse',
        metrics=[tf.keras.metrics.RecallAtPrecision(precision=0.8)])
    ```
    """

    @inject_mesh
    def __init__(
        self,
        recall,
        num_thresholds=200,
        class_id=None,
        name=None,
        dtype=None,
    ):
        if recall < 0 or recall > 1:
            raise ValueError(
                "Argument `recall` must be in the range [0, 1]. "
                f"Received: recall={recall}"
            )
        self.recall = recall
        self.num_thresholds = num_thresholds
        super().__init__(
            value=recall,
            num_thresholds=num_thresholds,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )

    def result(self):
        precisions = tf.math.divide_no_nan(
            self.true_positives,
            tf.math.add(self.true_positives, self.false_positives),
        )
        recalls = tf.math.divide_no_nan(
            self.true_positives,
            tf.math.add(self.true_positives, self.false_negatives),
        )
        return self._find_max_threshold_under_constraint(
            recalls, precisions, tf.greater_equal
        )

    def _find_max_threshold_under_constraint(self, constrained, dependent, predicate):
        """Returns the maximum of dependent_statistic that satisfies the
        constraint.

        Args:
          constrained: Over these values the constraint
            is specified. A rank-1 tensor.
          dependent: From these values the maximum that satiesfies the
            constraint is selected. Values in this tensor and in
            `constrained` are linked by having the same threshold at each
            position, hence this tensor must have the same shape.
          predicate: A binary boolean functor to be applied to arguments
          `constrained` and `self.value`, e.g. `tf.greater`.

        Returns:
          maximal dependent value, if no value satiesfies the constraint 0.0.
        """
        variable_dtype = self.true_positives.dtype
        thresholds = tf.convert_to_tensor(self.thresholds, dtype=variable_dtype)
        feasible = tf.where(predicate(constrained, self.value))
        feasible_exists = tf.greater(tf.size(feasible), 0)
        return tf.where(feasible_exists, tf.gather(thresholds, tf.gather(feasible, [0])), 0.0)

    def get_config(self):
        config = {
            "num_thresholds": self.num_thresholds,
            "recall": self.recall,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class ThresholdAtNPV(SensitivitySpecificityBase):
    """Computes best recall where precision is >= specified value.

    For a given score-label-distribution the required precision might not
    be achievable, in this case 0.0 is returned as recall.

    This metric creates four local variables, `true_positives`,
    `true_negatives`, `false_positives` and `false_negatives` that are used to
    compute the recall at the given precision. The threshold for the given
    precision value is computed and used to evaluate the corresponding recall.

    If `sample_weight` is `None`, weights default to 1.
    Use `sample_weight` of 0 to mask values.

    If `class_id` is specified, we calculate precision by considering only the
    entries in the batch for which `class_id` is above the threshold
    predictions, and computing the fraction of them for which `class_id` is
    indeed a correct label.

    Args:
      npv: A scalar value in range `[0, 1]`.
      num_thresholds: (Optional) Defaults to 200. The number of thresholds to
        use for matching the given precision.
      class_id: (Optional) Integer class ID for which we want binary metrics.
        This must be in the half-open interval `[0, num_classes)`, where
        `num_classes` is the last dimension of predictions.
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.RecallAtPrecision(0.8)
    >>> m.update_state([0, 0, 1, 1], [0, 0.5, 0.3, 0.9])
    >>> m.result().numpy()
    0.5

    >>> m.reset_state()
    >>> m.update_state([0, 0, 1, 1], [0, 0.5, 0.3, 0.9],
    ...                sample_weight=[1, 0, 0, 1])
    >>> m.result().numpy()
    1.0

    Usage with `compile()` API:

    ```python
    model.compile(
        optimizer='sgd',
        loss='mse',
        metrics=[tf.keras.metrics.RecallAtPrecision(precision=0.8)])
    ```
    """

    @inject_mesh
    def __init__(
        self,
        npv,
        num_thresholds=200,
        class_id=None,
        name=None,
        dtype=None,
    ):
        if npv < 0 or npv > 1:
            raise ValueError(
                "Argument `npv` must be in the range [0, 1]. "
                f"Received: npv={npv}"
            )
        self.npv = npv
        self.num_thresholds = num_thresholds
        super().__init__(
            value=npv,
            num_thresholds=num_thresholds,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )

    def result(self):
        npvs = tf.math.divide_no_nan(
            self.true_negatives,
            tf.math.add(self.true_negatives, self.false_negatives),
        )
        specificities = tf.math.divide_no_nan(
            self.true_negatives,
            tf.math.add(self.true_negatives, self.false_positives),
        )
        return self._find_max_threshold_under_constraint(
            npvs, specificities, tf.greater_equal
        )

    def _find_max_threshold_under_constraint(self, constrained, dependent, predicate):
        """Returns the maximum of dependent_statistic that satisfies the
        constraint.

        Args:
          constrained: Over these values the constraint
            is specified. A rank-1 tensor.
          dependent: From these values the maximum that satiesfies the
            constraint is selected. Values in this tensor and in
            `constrained` are linked by having the same threshold at each
            position, hence this tensor must have the same shape.
          predicate: A binary boolean functor to be applied to arguments
          `constrained` and `self.value`, e.g. `tf.greater`.

        Returns:
          maximal dependent value, if no value satiesfies the constraint 0.0.
        """
        variable_dtype = self.true_negatives.dtype
        thresholds = tf.convert_to_tensor(self.thresholds, dtype=variable_dtype)
        feasible = tf.where(predicate(constrained, self.value))
        feasible_exists = tf.greater(tf.size(feasible), 0)
        index = tf.size(feasible) - 1
        index = tf.where(tf.greater_equal(index, 0), index, 0)
        try:
            return tf.where(feasible_exists,
                            # tf.gather(thresholds, tf.gather(feasible, [tf.shape(feasible)[0] - 1])),
                            tf.gather(thresholds, tf.gather(feasible, [index])),
                            0.0)
        except Exception as e:
            print("Unexpected exception in thresholdAtNPV. " + str(e))
            return 0.0

    def get_config(self):
        config = {
            "num_thresholds": self.num_thresholds,
            "npv": self.npv,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


@tf.function
def tp_tf(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = y_true
    y_pred_label = y_pred
    real_positive_indexes_mask = tf.math.greater(y_true_label, 0)
    predicted_positive_indexes_mask = tf.math.greater(y_pred_label, threshold)
    positive_indexes_intersect_mask = tf.math.logical_and(predicted_positive_indexes_mask, real_positive_indexes_mask)
    predicted_positives = tf.cast(tf.size(tf.where(predicted_positive_indexes_mask)), tf.float32)
    predicted_positives_in_true_set = tf.cast(tf.size(tf.where(positive_indexes_intersect_mask)), tf.float32)
    tp = predicted_positives_in_true_set / (predicted_positives + tf.keras.backend.epsilon())
    return tp


@tf.function
def fp_tf(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = y_true
    y_pred_label = y_pred
    real_positive_indexes_mask = tf.math.greater(y_true_label, 0)
    predicted_positive_indexes_mask = tf.math.greater(y_pred_label, threshold)
    positive_indexes_intersect_mask = tf.math.logical_and(predicted_positive_indexes_mask, real_positive_indexes_mask)
    predicted_positives = tf.cast(tf.size(tf.where(predicted_positive_indexes_mask)), tf.float32)
    predicted_positives_in_true_set = tf.cast(tf.size(tf.where(positive_indexes_intersect_mask)), tf.float32)
    fp = (predicted_positives - predicted_positives_in_true_set) / (predicted_positives + K.epsilon())
    return fp


@tf.function
def fn_tf(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    real_negative_indexes_mask = tf.math.equal(y_true_label, 0)
    predicted_negative_indexes_mask = tf.math.less_equal(y_pred_label, threshold)
    negative_indexes_intersect_mask = tf.math.logical_and(predicted_negative_indexes_mask, real_negative_indexes_mask)
    predicted_negative_indexes_mask = tf.math.less_equal(y_pred_label, threshold)
    predicted_negatives = tf.cast(tf.size(tf.where(predicted_negative_indexes_mask)), tf.float32)
    predicted_negatives_in_true_set = tf.cast(tf.size(tf.where(negative_indexes_intersect_mask)), tf.float32)
    fn = (predicted_negatives - predicted_negatives_in_true_set) / (predicted_negatives + K.epsilon())
    return fn


@tf.function
def tn_tf(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    real_negative_indexes_mask = tf.math.equal(y_true_label, 0)
    predicted_negative_indexes_mask = tf.math.less_equal(y_pred_label, threshold)
    negative_indexes_intersect_mask = tf.math.logical_and(predicted_negative_indexes_mask, real_negative_indexes_mask)
    predicted_negative_indexes_mask = tf.math.less_equal(y_pred_label, threshold)
    predicted_negatives = tf.cast(tf.size(tf.where(predicted_negative_indexes_mask)), tf.float32)
    predicted_negatives_in_true_set = tf.cast(tf.size(tf.where(negative_indexes_intersect_mask)), tf.float32)
    tn = predicted_negatives_in_true_set / (predicted_negatives + tf.keras.backend.epsilon())
    return tn


@tf.function
def precision_tf(y_true, y_pred, label_index=0, threshold=0.5):
    tp = tp_tf(y_true, y_pred, label_index, threshold)
    fp = fp_tf(y_true, y_pred, label_index, threshold)
    return tp / (tp + fp + K.epsilon())


@tf.function
def recall_tf(y_true, y_pred, label_index=0, threshold=0.5):
    tp = tp_tf(y_true, y_pred, label_index, threshold)
    fn = fn_tf(y_true, y_pred, label_index, threshold)
    return tp / (tp + fn + K.epsilon())


@tf.function
def f1_score_tf(y_true, y_pred, label_index=0, threshold=0.5):
    precision = precision_tf(y_true, y_pred, label_index, threshold)
    recall = recall_tf(y_true, y_pred, label_index, threshold)
    return 2 * precision * recall / (precision + recall + K.epsilon())

@tf.function
def precision_at_k(y_true, y_pred, label_index=0, k=100, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    values, indices = tf.math.top_k(y_pred_label, k=k)
    y_true_label = tf.gather(y_true_label, indices)
    y_pred_label = tf.gather(y_pred_label, indices)
    return precision_tf(y_true_label, y_pred_label, label_index=0, threshold=threshold)

@tf.function
def mean_positive_value(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    positive_pred_indexes = tf.where(tf.greater(y_pred_label, threshold))
    return tf.reduce_mean(tf.gather(y_pred_label, positive_pred_indexes))

@tf.function
def mean_true_positive_value(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    tp_pred_indexes = tf.where(tf.logical_and(tf.greater(y_pred_label, threshold), tf.equal(y_true_label, 1)))
    return tf.reduce_mean(tf.gather(y_pred_label, tp_pred_indexes))

@tf.function
def mean_false_positive_value(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    fp_pred_indexes = tf.where(tf.logical_and(tf.greater(y_pred_label, threshold), tf.equal(y_true_label, 0)))
    return tf.reduce_mean(tf.gather(y_pred_label, fp_pred_indexes))

@tf.function
def mean_negative_value(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    negative_pred_indexes = tf.where(tf.less(y_pred_label, threshold))
    return tf.reduce_mean(tf.gather(y_pred_label, negative_pred_indexes))

@tf.function
def mean_true_negative_value(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    tp_pred_indexes = tf.where(tf.logical_and(tf.less(y_pred_label, threshold), tf.equal(y_true_label, 0)))
    return tf.reduce_mean(tf.gather(y_pred_label, tp_pred_indexes))

@tf.function
def mean_false_negative_value(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_true_label = y_true_label[:, 0]
    y_pred_label = y_pred_label[:, 0]
    tp_pred_indexes = tf.where(tf.logical_and(tf.less(y_pred_label, threshold), tf.equal(y_true_label, 1)))
    return tf.reduce_mean(tf.gather(y_pred_label, tp_pred_indexes))

@tf.function
def confusion_matrix_values_tf(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = tf.gather(y_true, [label_index], axis=1)
    y_pred_label = tf.gather(y_pred, [label_index], axis=1)
    y_pred_label = tf.reshape(y_pred_label, tf.shape(y_pred_label)[0])
    y_true_label = tf.reshape(y_true_label, tf.shape(y_true_label)[0])
    real_positive_indexes_mask = tf.math.greater(y_true_label, 0)
    predicted_positive_indexes_mask = tf.math.greater(y_pred_label, threshold)
    positive_indexes_intersect_mask = tf.math.logical_and(predicted_positive_indexes_mask, real_positive_indexes_mask)
    predicted_positives = tf.where(predicted_positive_indexes_mask).shape.as_list()[0]
    predicted_positives_in_true_set = tf.where(positive_indexes_intersect_mask).shape.as_list()[0]
    tp = predicted_positives_in_true_set / (predicted_positives + tf.keras.backend.epsilon())
    real_negative_indexes_mask = tf.math.equal(y_true_label, 0)
    predicted_negative_indexes_mask = tf.math.less_equal(y_pred_label, threshold)
    negative_indexes_intersect_mask = tf.math.logical_and(predicted_negative_indexes_mask, real_negative_indexes_mask)
    predicted_negatives = tf.where(predicted_negative_indexes_mask).shape.as_list()[0]
    predicted_negatives_in_true_set = tf.where(negative_indexes_intersect_mask).shape.as_list()[0]
    correct_predictions = predicted_positives_in_true_set + predicted_negatives_in_true_set
    total_predictions = y_true.shape.as_list()[0]
    accuracy = correct_predictions / (total_predictions + K.epsilon())
    error_rate = 1 - accuracy
    #TODO decide what values to be returned when no positives are detected
    if predicted_positives == 0:
        return [[np.nan, np.nan], [np.nan, np.nan]], accuracy, error_rate, np.nan, np.nan, np.nan, np.nan
    else:
        tn = predicted_negatives_in_true_set / (predicted_negatives + tf.keras.backend.epsilon())
        fp = (predicted_positives - predicted_positives_in_true_set) / (predicted_positives + K.epsilon())
        fn = (predicted_negatives - predicted_negatives_in_true_set) / (predicted_negatives + K.epsilon())
        precision = tp / (tp + fp + K.epsilon())
        recall = tp / (tp + fn + K.epsilon())
        f1_score = 2 * precision * recall / (precision + recall + K.epsilon())
        return [[tp, fp], [fn, tn]], accuracy, error_rate, precision, recall, f1_score, np.nan


def confusion_matrix_values(y_true, y_pred, label_index=0, threshold=0.5):
    y_true_label = np.transpose(y_true)[label_index]
    y_pred_label = np.transpose(y_pred)[label_index]
    real_positive_indexes_mask = y_true_label > 0
    predicted_positive_indexes_mask = y_pred_label > threshold
    positive_indexes_intersect_mask = np.logical_and(predicted_positive_indexes_mask, real_positive_indexes_mask)
    predicted_positives = len(np.argwhere(predicted_positive_indexes_mask).flatten())
    predicted_positives_in_true_set = len(np.argwhere(positive_indexes_intersect_mask).flatten())
    tp = predicted_positives_in_true_set / (predicted_positives + tf.keras.backend.epsilon())
    real_negative_indexes_mask = y_true_label == 0
    predicted_negative_indexes_mask = y_pred_label < threshold
    negative_indexes_intersect_mask = np.logical_and(predicted_negative_indexes_mask, real_negative_indexes_mask)
    predicted_negatives = len(np.argwhere(predicted_negative_indexes_mask).flatten())
    predicted_negatives_in_true_set = len(np.argwhere(negative_indexes_intersect_mask).flatten())
    correct_predictions = predicted_positives_in_true_set + predicted_negatives_in_true_set
    total_predictions = len(y_true)
    accuracy = correct_predictions / (total_predictions + K.epsilon())
    error_rate = 1 - accuracy
    #TODO decide what values to be returned when no positives are detected
    if predicted_positives == 0:
        return [[np.nan, np.nan], [np.nan, np.nan]], accuracy, error_rate, np.nan, np.nan, np.nan
    else:
        tn = predicted_negatives_in_true_set / (predicted_negatives + K.epsilon())
        fp = (predicted_positives - predicted_positives_in_true_set) / (predicted_positives + K.epsilon())
        fn = (predicted_negatives - predicted_negatives_in_true_set) / (predicted_negatives + K.epsilon())
        precision = tp / (tp + fp + K.epsilon())
        recall = tp / (tp + fn + K.epsilon())
        f1_score = 2 * precision * recall / (precision + recall + K.epsilon())
        return [[tp, fp], [fn, tn]], accuracy, error_rate, precision, recall, f1_score

def area_under_curve(y_true, y_pred, label_index=0, thresholds=100, method='roc'):
    confusion_matrixes = np.empty((thresholds, 2, 2))
    accuracies = np.zeros(thresholds)
    error_rates = np.zeros(thresholds)
    precisions = np.zeros(thresholds)
    recalls = np.zeros(thresholds)
    f1_scores = np.zeros(thresholds)
    threshold_values = np.flip(np.linspace(0, 1, thresholds))
    for index, threshold in enumerate(threshold_values):
        confusion_matrixes[index], accuracies[index], error_rates[index], precisions[index], recalls[index], \
            f1_scores[index] = confusion_matrix_values(y_true, y_pred, label_index=label_index, threshold=threshold)
    nan_values = np.logical_and(np.isnan(recalls), np.isnan(precisions))
    x = recalls[~nan_values]
    y = precisions[~nan_values]
    threshold_values = threshold_values[~nan_values]
    return x, y, threshold_values, np.trapz(y, x)


class CategoricalClassMetric(tf.keras.metrics.Metric):
    def __init__(self, class_index, metric_function, batch_size, name="categorical_class_metric", metric_threshold=0.5,
                 **kwargs):
        super(CategoricalClassMetric, self).__init__(name=name, **kwargs)
        self.batch_size = batch_size
        self.metric_function = metric_function
        self.metric_threshold = metric_threshold
        self.class_index = class_index
        self.metric = 0

    def update_state(self, y_true, y_pred, sample_weight=None):
        self.metric = self.metric_function(y_true, y_pred, self.class_index, self.metric_threshold)

    def result(self):
        return self.metric

class NegativePredictiveValue(Metric):
    """Computes the NPV of the predictions with respect to the labels.

    The metric creates two local variables, `true_positives` and
    `false_positives` that are used to compute the precision. This value is
    ultimately returned as `precision`, an idempotent operation that simply
    divides `true_positives` by the sum of `true_positives` and
    `false_positives`.

    If `sample_weight` is `None`, weights default to 1.
    Use `sample_weight` of 0 to mask values.

    If `top_k` is set, we'll calculate precision as how often on average a class
    among the top-k classes with the highest predicted values of a batch entry
    is correct and can be found in the label for that entry.

    If `class_id` is specified, we calculate precision by considering only the
    entries in the batch for which `class_id` is above the threshold and/or in
    the top-k highest predictions, and computing the fraction of them for which
    `class_id` is indeed a correct label.

    Args:
      thresholds: (Optional) A float value, or a Python list/tuple of float
        threshold values in [0, 1]. A threshold is compared with prediction
        values to determine the truth value of predictions (i.e., above the
        threshold is `true`, below is `false`). If used with a loss function
        that sets `from_logits=True` (i.e. no sigmoid applied to predictions),
        `thresholds` should be set to 0. One metric value is generated for each
        threshold value. If neither thresholds nor top_k are set, the default is
        to calculate precision with `thresholds=0.5`.
      top_k: (Optional) Unset by default. An int value specifying the top-k
        predictions to consider when calculating precision.
      class_id: (Optional) Integer class ID for which we want binary metrics.
        This must be in the half-open interval `[0, num_classes)`, where
        `num_classes` is the last dimension of predictions.
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.Precision()
    >>> m.update_state([0, 1, 1, 1], [1, 0, 1, 1])
    >>> m.result().numpy()
    0.6666667

    >>> m.reset_state()
    >>> m.update_state([0, 1, 1, 1], [1, 0, 1, 1], sample_weight=[0, 0, 1, 0])
    >>> m.result().numpy()
    1.0

    >>> # With top_k=2, it will calculate precision over y_true[:2]
    >>> # and y_pred[:2]
    >>> m = tf.keras.metrics.Precision(top_k=2)
    >>> m.update_state([0, 0, 1, 1], [1, 1, 1, 1])
    >>> m.result().numpy()
    0.0

    >>> # With top_k=4, it will calculate precision over y_true[:4]
    >>> # and y_pred[:4]
    >>> m = tf.keras.metrics.Precision(top_k=4)
    >>> m.update_state([0, 0, 1, 1], [1, 1, 1, 1])
    >>> m.result().numpy()
    0.5

    Usage with `compile()` API:

    ```python
    model.compile(optimizer='sgd',
                  loss='binary_crossentropy',
                  metrics=[tf.keras.metrics.Precision()])
    ```

    Usage with a loss with `from_logits=True`:

    ```python
    model.compile(optimizer='adam',
                  loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
                  metrics=[tf.keras.metrics.Precision(thresholds=0)])
    ```
    """

    @inject_mesh
    def __init__(
        self, thresholds=None, top_k=None, class_id=None, name=None, dtype=None
    ):
        super().__init__(name=name, dtype=dtype)
        self.init_thresholds = thresholds
        self.top_k = top_k
        self.class_id = class_id

        default_threshold = 0.5 if top_k is None else metrics_utils.NEG_INF
        self.thresholds = metrics_utils.parse_init_thresholds(
            thresholds, default_threshold=default_threshold
        )
        self._thresholds_distributed_evenly = (
            metrics_utils.is_evenly_distributed_thresholds(self.thresholds)
        )
        self.true_negatives = self.add_weight(
            name="true_negatives", shape=(len(self.thresholds),), initializer="zeros"
        )
        self.false_negatives = self.add_weight(
            name="false_negatives",
            shape=(len(self.thresholds),),
            initializer="zeros",
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        """Accumulates true positive and false positive statistics.

        Args:
          y_true: The ground truth values, with the same dimensions as `y_pred`.
            Will be cast to `bool`.
          y_pred: The predicted values. Each element must be in the range
            `[0, 1]`.
          sample_weight: Optional weighting of each example. Can
            be a `Tensor` whose rank is either 0, or the same rank as `y_true`,
            and must be broadcastable to `y_true`. Defaults to `1`.

        Returns:
          Update op.
        """
        return metrics_utils.update_confusion_matrix_variables(
            {
                metrics_utils.ConfusionMatrix.TRUE_NEGATIVES: self.true_negatives,  # noqa: E501
                metrics_utils.ConfusionMatrix.FALSE_NEGATIVES: self.false_negatives,  # noqa: E501
            },
            y_true,
            y_pred,
            thresholds=self.thresholds,
            thresholds_distributed_evenly=self._thresholds_distributed_evenly,
            top_k=self.top_k,
            class_id=self.class_id,
            sample_weight=sample_weight,
        )

    def result(self):
        result = tf.math.divide_no_nan(
            self.true_negatives,
            tf.math.add(self.true_negatives, self.false_negatives),
        )
        return result[0] if len(self.thresholds) == 1 else result

    def reset_state(self):
        num_thresholds = len(to_list(self.thresholds))
        backend.batch_set_value(
            [
                (v, np.zeros((num_thresholds,)))
                for v in (self.true_negatives, self.false_negatives)
            ]
        )

    def get_config(self):
        config = {
            "thresholds": self.init_thresholds,
            "top_k": self.top_k,
            "class_id": self.class_id,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class Specificity(Metric):
    """Computes the NPV of the predictions with respect to the labels.

    The metric creates two local variables, `true_positives` and
    `false_positives` that are used to compute the precision. This value is
    ultimately returned as `precision`, an idempotent operation that simply
    divides `true_positives` by the sum of `true_positives` and
    `false_positives`.

    If `sample_weight` is `None`, weights default to 1.
    Use `sample_weight` of 0 to mask values.

    If `top_k` is set, we'll calculate precision as how often on average a class
    among the top-k classes with the highest predicted values of a batch entry
    is correct and can be found in the label for that entry.

    If `class_id` is specified, we calculate precision by considering only the
    entries in the batch for which `class_id` is above the threshold and/or in
    the top-k highest predictions, and computing the fraction of them for which
    `class_id` is indeed a correct label.

    Args:
      thresholds: (Optional) A float value, or a Python list/tuple of float
        threshold values in [0, 1]. A threshold is compared with prediction
        values to determine the truth value of predictions (i.e., above the
        threshold is `true`, below is `false`). If used with a loss function
        that sets `from_logits=True` (i.e. no sigmoid applied to predictions),
        `thresholds` should be set to 0. One metric value is generated for each
        threshold value. If neither thresholds nor top_k are set, the default is
        to calculate precision with `thresholds=0.5`.
      top_k: (Optional) Unset by default. An int value specifying the top-k
        predictions to consider when calculating precision.
      class_id: (Optional) Integer class ID for which we want binary metrics.
        This must be in the half-open interval `[0, num_classes)`, where
        `num_classes` is the last dimension of predictions.
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.Precision()
    >>> m.update_state([0, 1, 1, 1], [1, 0, 1, 1])
    >>> m.result().numpy()
    0.6666667

    >>> m.reset_state()
    >>> m.update_state([0, 1, 1, 1], [1, 0, 1, 1], sample_weight=[0, 0, 1, 0])
    >>> m.result().numpy()
    1.0

    >>> # With top_k=2, it will calculate precision over y_true[:2]
    >>> # and y_pred[:2]
    >>> m = tf.keras.metrics.Precision(top_k=2)
    >>> m.update_state([0, 0, 1, 1], [1, 1, 1, 1])
    >>> m.result().numpy()
    0.0

    >>> # With top_k=4, it will calculate precision over y_true[:4]
    >>> # and y_pred[:4]
    >>> m = tf.keras.metrics.Precision(top_k=4)
    >>> m.update_state([0, 0, 1, 1], [1, 1, 1, 1])
    >>> m.result().numpy()
    0.5

    Usage with `compile()` API:

    ```python
    model.compile(optimizer='sgd',
                  loss='binary_crossentropy',
                  metrics=[tf.keras.metrics.Precision()])
    ```

    Usage with a loss with `from_logits=True`:

    ```python
    model.compile(optimizer='adam',
                  loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
                  metrics=[tf.keras.metrics.Precision(thresholds=0)])
    ```
    """

    @inject_mesh
    def __init__(
        self, thresholds=None, top_k=None, class_id=None, name=None, dtype=None
    ):
        super().__init__(name=name, dtype=dtype)
        self.init_thresholds = thresholds
        self.top_k = top_k
        self.class_id = class_id

        default_threshold = 0.5 if top_k is None else metrics_utils.NEG_INF
        self.thresholds = metrics_utils.parse_init_thresholds(
            thresholds, default_threshold=default_threshold
        )
        self._thresholds_distributed_evenly = (
            metrics_utils.is_evenly_distributed_thresholds(self.thresholds)
        )
        self.true_negatives = self.add_weight(
            name="true_negatives", shape=(len(self.thresholds),), initializer="zeros"
        )
        self.false_positives = self.add_weight(
            name="false_positives",
            shape=(len(self.thresholds),),
            initializer="zeros",
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        """Accumulates true positive and false positive statistics.

        Args:
          y_true: The ground truth values, with the same dimensions as `y_pred`.
            Will be cast to `bool`.
          y_pred: The predicted values. Each element must be in the range
            `[0, 1]`.
          sample_weight: Optional weighting of each example. Can
            be a `Tensor` whose rank is either 0, or the same rank as `y_true`,
            and must be broadcastable to `y_true`. Defaults to `1`.

        Returns:
          Update op.
        """
        return metrics_utils.update_confusion_matrix_variables(
            {
                metrics_utils.ConfusionMatrix.TRUE_NEGATIVES: self.true_negatives,  # noqa: E501
                metrics_utils.ConfusionMatrix.FALSE_POSITIVES: self.false_positives,  # noqa: E501
            },
            y_true,
            y_pred,
            thresholds=self.thresholds,
            thresholds_distributed_evenly=self._thresholds_distributed_evenly,
            top_k=self.top_k,
            class_id=self.class_id,
            sample_weight=sample_weight,
        )

    def result(self):
        result = tf.math.divide_no_nan(
            self.true_negatives,
            tf.math.add(self.true_negatives, self.false_positives),
        )
        return result[0] if len(self.thresholds) == 1 else result

    def reset_state(self):
        num_thresholds = len(to_list(self.thresholds))
        backend.batch_set_value(
            [
                (v, np.zeros((num_thresholds,)))
                for v in (self.true_negatives, self.false_positives)
            ]
        )

    def get_config(self):
        config = {
            "thresholds": self.init_thresholds,
            "top_k": self.top_k,
            "class_id": self.class_id,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class SpecificityAtNPV(SensitivitySpecificityBase):
    """Computes best recall where precision is >= specified value.

    For a given score-label-distribution the required precision might not
    be achievable, in this case 0.0 is returned as recall.

    This metric creates four local variables, `true_positives`,
    `true_negatives`, `false_positives` and `false_negatives` that are used to
    compute the recall at the given precision. The threshold for the given
    precision value is computed and used to evaluate the corresponding recall.

    If `sample_weight` is `None`, weights default to 1.
    Use `sample_weight` of 0 to mask values.

    If `class_id` is specified, we calculate precision by considering only the
    entries in the batch for which `class_id` is above the threshold
    predictions, and computing the fraction of them for which `class_id` is
    indeed a correct label.

    Args:
      npv: A scalar value in range `[0, 1]`.
      num_thresholds: (Optional) The number of thresholds to
        use for matching the given precision. Defaults to `200`.
      class_id: (Optional) Integer class ID for which we want binary metrics.
        This must be in the half-open interval `[0, num_classes)`, where
        `num_classes` is the last dimension of predictions.
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.RecallAtPrecision(0.8)
    >>> m.update_state([0, 0, 1, 1], [0, 0.5, 0.3, 0.9])
    >>> m.result().numpy()
    0.5

    >>> m.reset_state()
    >>> m.update_state([0, 0, 1, 1], [0, 0.5, 0.3, 0.9],
    ...                sample_weight=[1, 0, 0, 1])
    >>> m.result().numpy()
    1.0

    Usage with `compile()` API:

    ```python
    model.compile(
        optimizer='sgd',
        loss='binary_crossentropy',
        metrics=[tf.keras.metrics.RecallAtPrecision(precision=0.8)])
    ```
    """

    @inject_mesh
    def __init__(
        self,
        npv,
        num_thresholds=200,
        class_id=None,
        name=None,
        dtype=None,
    ):
        if npv < 0 or npv > 1:
            raise ValueError(
                "Argument `npv` must be in the range [0, 1]. "
                f"Received: npv={npv}"
            )
        self.npv = npv
        self.num_thresholds = num_thresholds
        super().__init__(
            value=npv,
            num_thresholds=num_thresholds,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )

    def result(self):
        npvs = tf.math.divide_no_nan(
            self.true_negatives,
            tf.math.add(self.true_negatives, self.false_negatives),
        )
        sensitivities = tf.math.divide_no_nan(
            self.true_negatives,
            tf.math.add(self.true_negatives, self.false_positives),
        )
        return self._find_max_under_constraint(
            npvs, sensitivities, tf.greater_equal
        )

    def get_config(self):
        config = {
            "num_thresholds": self.num_thresholds,
            "npv": self.npv,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



import os

import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.utils import shuffle
from tensorflow.keras.utils import Sequence

class SantoPredictGenerator(Sequence):
    def __init__(self, times, lcs, input_size=500, step_size=1, batch_size=500, zero_epsilon=1e-7):
        self.times = times
        self.lcs = lcs
        self.input_size = input_size
        self.step_size = step_size
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.zero_epsilon = zero_epsilon
        self.steps_count, self.generator_info_df = self._count_inputs()

    def _count_inputs(self):
        generator_info_df: pd.DataFrame = pd.DataFrame(columns=['filename', 'file_index', 'batch_index'])
        lcs_to_original_lcs: list[tuple[int, int, int, list[float]]] = []
        for curve_index, times in enumerate(self.times):
            diff = np.diff(times)
            jump_indices = np.where(abs(diff) > 12 / 24)[0]
            previous_jump_index = 0
            jump_index = len(self.lcs[curve_index])
            if len(jump_indices) == 0:
                lcs_to_original_lcs = lcs_to_original_lcs + [(curve_index, previous_jump_index, jump_index, self.lcs[curve_index])]
            else:
                for jump_index in jump_indices:
                    lcs_to_original_lcs = lcs_to_original_lcs + [(curve_index, previous_jump_index, jump_index, self.lcs[curve_index][previous_jump_index:jump_index + 1])]
                    previous_jump_index = jump_index + 1
                if jump_index < len(self.lcs[curve_index]) - 1:
                    jump_index = len(self.lcs[curve_index]) - 1
                    lcs_to_original_lcs = lcs_to_original_lcs + [(curve_index, previous_jump_index, jump_index, self.lcs[curve_index][previous_jump_index:jump_index + 1])]
        batch_index = 0
        for lc_to_original_lc in lcs_to_original_lcs:
            flux = lc_to_original_lc[3]
            file_batch_indexes = np.arange(0, len(flux), self.batch_size)
            for file_index in file_batch_indexes:
                generator_info_df = pd.concat([generator_info_df, pd.DataFrame.from_dict(
                                    {'curve_index': [lc_to_original_lc[0]], 'initial_jump_index': [lc_to_original_lc[1]],
                                     'end_jump_index': [lc_to_original_lc[2]], 'file_index': [file_index],
                                     'batch_index': [batch_index]})], ignore_index=True)
                batch_index = batch_index + 1
        return batch_index - 1 , generator_info_df

    def __len__(self):
        return len(self.generator_info_df['batch_index'])

    def __getitem__(self, index):
        curve_index: int = int(self.generator_info_df.iloc[index]['curve_index'])
        initial_jump_index: int = int(self.generator_info_df.iloc[index]['initial_jump_index'])
        end_jump_index: int = int(self.generator_info_df.iloc[index]['end_jump_index'])
        file_index: int = int(self.generator_info_df.iloc[index]['file_index'])
        flux = self.lcs[curve_index][initial_jump_index:end_jump_index + 1]
        max_index = np.min([len(flux) - self.input_size, file_index + self.batch_size])
        train_fluxes = np.full((self.batch_size,  self.input_size), self.zero_epsilon)
        train_tags = np.full((self.batch_size, 1), 0)
        for iteration_index in np.arange(file_index, max_index):
            flux_data = flux[iteration_index:iteration_index + self.input_size] / 2
            train_fluxes[iteration_index - file_index] = flux_data
        return train_fluxes, None

    def on_epoch_end(self):
        if self.shuffle:
            self.generator_info_df = shuffle(self.generator_info_df)

    def class_weights(self):
        return {0: 1, 1: 1}

    def steps_per_epoch(self):
        return self.steps_count


import numpy as np


def binary_encode_times(times, length):
    times_binary_str = np.binary_repr(times)
    return [np.array(''.split(times_binary_str.zfill(length))).astype(int) for time_binary_str in times_binary_str]


def value_encode_times(times, max_time=1000):
    return times % max_time / max_time


import logging

from keras.callbacks import EarlyStopping
import numpy as np


class ExoMlEarlyStopping(EarlyStopping):
    def on_train_end(self, logs=None):
        super().on_train_end(logs)
        if self.stopped_epoch > 0 and self.verbose > 0:
            logging.warning('Epoch %05d: early stopping', (self.stopped_epoch + 1))

    def on_train_batch_begin(self, batch, logs=None):
        if np.isnan(self.model.loss):
            raise ValueError("Loss is NaN")
        return super().on_train_batch_begin(batch, logs)



import glob
import gzip
import multiprocessing
import os
import pathlib
import re
import shutil
from multiprocessing import Pool
import pandas as pd
import numpy as np

from exoml.preparation.parser_utils import CreateTargetInput, create_target_csv, create_triceratops_prob

def create_targets_df(tces_dir):
    # Merge TCEs DF and cumulative DF ignoring everything not from Kepler or K2
    # Mark TCEs as planet only if they have photometric and rv measurements
    # test_df = pd.read_csv(tces_dir + '/classified_tces.csv', comment='#')
    # test_df = test_df.sort_values(by=["object_id", "period"], ascending=True)
    # test_df = test_df[test_df['type'].isin(['fp', 'planet', 'planet_transit', 'tce'])]
    # test_df.to_csv(tces_dir + '/injected_objects_tces.csv')
    cfps_df = pd.read_csv(tces_dir + '/fpwg_2023.04.25_10.26.55.csv', comment='#')
    kois_df = pd.read_csv(tces_dir + '/cumulative_2023.06.30_13.09.52.csv', comment='#')
                          #usecols=['rowid','kepid','kepoi_name','kepler_name','koi_disposition','koi_vet_stat','koi_vet_date','koi_pdisposition','koi_score','koi_fpflag_nt','koi_fpflag_ss','koi_fpflag_co','koi_fpflag_ec','koi_disp_prov','koi_comment','koi_period','koi_time0bk','koi_time0','koi_eccen','koi_longp','koi_impact','koi_duration','koi_ingress','koi_depth','koi_ror','koi_srho','koi_fittype','koi_prad','koi_sma','koi_incl','koi_teq','koi_insol','koi_dor','koi_limbdark_mod','koi_ldm_coeff4','koi_ldm_coeff3','koi_ldm_coeff2','koi_ldm_coeff1','koi_parm_prov','koi_max_sngle_ev','koi_max_mult_ev','koi_model_snr','koi_count','koi_num_transits','koi_tce_plnt_num','koi_tce_delivname','koi_quarters','koi_bin_oedp_sig','koi_trans_mod','koi_model_dof','koi_model_chisq','koi_datalink_dvr','koi_datalink_dvs','koi_steff','koi_slogg','koi_smet','koi_srad','koi_smass','koi_sage','koi_sparprov','ra','dec','koi_kepmag','koi_gmag','koi_rmag','koi_imag','koi_zmag','koi_jmag','koi_hmag','koi_kmag','koi_fwm_stat_sig','koi_fwm_sra','koi_fwm_sdec','koi_fwm_srao','koi_fwm_sdeco','koi_fwm_prao','koi_fwm_pdeco','koi_dicco_mra','koi_dicco_mdec','koi_dicco_msky','koi_dikco_mra','koi_dikco_mdec','koi_dikco_msky'])
    tces_df = pd.read_csv(tces_dir + '/q1_q17_dr25_tce_2023.07.03_23.17.12.csv', comment='#')
    disc_df = pd.read_csv(tces_dir + '/PSCompPars_2023.06.30_07.30.12.csv', comment='#')
                          #usecols=['pl_name','hostname','tic_id','sy_snum','sy_pnum','discoverymethod','disc_year','disc_facility','rv_flag','tran_flag','pl_controv_flag','pl_orbper','pl_orbsmax','pl_rade','pl_radj','pl_bmasse','pl_bmassj','pl_bmassprov','pl_orbeccen','pl_insol','pl_eqt','ttv_flag','st_spectype','st_teff','st_rad','st_mass','st_met','st_metratio','st_logg','rastr','ra','decstr','dec','sy_dist','sy_vmag','sy_jmag','sy_hmag','sy_kmag','sy_gaiamag','sy_tmag','sy_kepmag','st_nphot','st_nrvc', 'disc_refname'])
    tces_df = tces_df[tces_df['tce_rogue_flag'] == 0]
    kois_df = kois_df[~((kois_df['koi_disposition'] == 'CONFIRMED') & (kois_df['koi_pdisposition'] == 'FALSE POSITIVE'))]
    tces_df['cent_so'] = (tces_df['tce_fwm_srao'] ** 2 + tces_df['tce_fwm_sdeco'] ** 2) ** 0.5
    tces_df['cent_so_err'] = (((tces_df['tce_fwm_srao'] / np.sqrt(tces_df['tce_fwm_srao'] ** 2 + tces_df['tce_fwm_sdeco'] ** 2)) * tces_df['tce_fwm_srao_err']) ** 2 + \
                              ((tces_df['tce_fwm_sdeco'] / np.sqrt(tces_df['tce_fwm_srao'] ** 2 + tces_df['tce_fwm_sdeco'] ** 2)) * tces_df['tce_fwm_sdeco_err']) ** 2) ** 0.5
    tces_df['cent_so_sigma'] = tces_df['cent_so'] / tces_df['cent_so_err']
    tces_df['source_offset_sigma'] = tces_df['tce_dicco_msky'] / tces_df['tce_dicco_msky_err']
    tces_df['kic_offset_sigma'] = tces_df['tce_dikco_msky'] / tces_df['tce_dikco_msky_err']
    # tces_df[~(((tces_df['tce_hap_stat'] / tces_df['tce_cap_stat'] > 1) |
    #            (tces_df['tce_hap_stat'] > 0) & (tces_df['tce_cap_stat'] < 0)) |
    #           (tces_df['tce_maxmes'] > 7.1) |
    #           (tces_df['cent_so_sigma'] > 3) |
    #           (tces_df['source_offset_sigma'] > 3) |
    #           (tces_df['kic_offset_sigma'] > 3)) & (tces_df['tce_model_snr'] > 3.5)]
    kois_df = pd.merge(kois_df, cfps_df, on=['kepoi_name'], how="outer", indicator=True)
    kois_df = kois_df[(kois_df['_merge'] == 'left_only') | (kois_df['_merge'] == 'both')]
    kois_df['kepid'] = kois_df['kepid_x']
    kois_df['pl_name'] = kois_df['kepler_name']
    kois_df['tce_period'] = kois_df['koi_period']
    kois_df['tce_period_round'] = kois_df['tce_period'].round(1)
    tces_df['tce_period_round'] = tces_df['tce_period'].round(1)
    tces_df = pd.merge(tces_df, kois_df, on=['kepid', 'tce_period_round'], how='outer')
    disc_df = disc_df[disc_df['disc_facility'] == 'Kepler']
    disc_df = disc_df[disc_df['discoverymethod'] == 'Transit']
    disc_df = pd.merge(tces_df, disc_df, on='pl_name', how='outer')
    disc_df['close_period'] = 0
    disc_df['type'] = 'tce'
    disc_df.loc[~disc_df['pl_name'].isna(), 'type'] = 'planet'
    disc_df.loc[(disc_df['koi_disposition'] == 'FALSE POSITIVE') & (~disc_df['type'].str.contains('planet')), 'type'] = 'fp_candidate'
    disc_df.loc[(disc_df['koi_disposition'] == 'CONFIRMED'), 'type'] = 'planet_transit'
    disc_df.loc[(disc_df['rv_flag'] > 0) & (disc_df['tran_flag'] > 0) & (disc_df['koi_disposition'] == 'CONFIRMED'), 'type'] = 'planet'
    disc_df.loc[(disc_df['rv_flag'] == 0) & (disc_df['tran_flag'] > 0) & (disc_df['koi_disposition'] == 'CONFIRMED'), 'type'] = 'planet_transit'
    disc_df.loc[(disc_df['koi_disposition'] == 'FALSE POSITIVE') & (~disc_df['type'].str.contains('planet')), 'type'] = 'fp_candidate'
    disc_df.loc[(disc_df['koi_disposition'] == 'CANDIDATE') & (~disc_df['type'].str.contains('planet')), 'type'] = 'candidate'
    disc_df.loc[(disc_df['fpwg_disp_status'] == 'CERTIFIED FP') & (~disc_df['type'].str.contains('planet')), 'type'] = 'fp'
    disc_df.loc[(disc_df['fpwg_disp_status'] == 'CERTIFIED FA') & (~disc_df['type'].str.contains('planet')), 'type'] = 'fa'
    disc_df['object_id'] = 'KIC ' + disc_df['kepid'].astype('Int64').astype('str')
    disc_df['period'] = disc_df['tce_period_x']
    disc_df.loc[disc_df['period'].isna(), 'period'] = disc_df[disc_df['period'].isna()]['tce_period_y']
    disc_df.loc[disc_df['period'].isna(), 'period'] = disc_df[disc_df['period'].isna()]['pl_orbper']
    disc_df['epoch'] = disc_df['tce_time0bk']
    disc_df.loc[disc_df['epoch'].isna(), 'epoch'] = disc_df[disc_df['epoch'].isna()]['koi_time0bk']
    disc_df.loc[disc_df['epoch'].isna(), 'epoch'] = disc_df[disc_df['epoch'].isna()]['pl_tranmid'] - 2454833.0
    disc_df['duration(h)'] = disc_df['pl_trandur']
    disc_df.loc[disc_df['duration(h)'].isna(), 'duration(h)'] = disc_df[disc_df['duration(h)'].isna()]['koi_duration']
    disc_df.loc[disc_df['duration(h)'].isna(), 'duration(h)'] = disc_df[disc_df['duration(h)'].isna()]['tce_duration']
    disc_df['depth_primary'] = disc_df['pl_trandep'] * 10000
    disc_df.loc[disc_df['depth_primary'].isna(), 'depth_primary'] = disc_df[disc_df['depth_primary'].isna()]['koi_depth']
    disc_df.loc[disc_df['depth_primary'].isna(), 'depth_primary'] = disc_df[disc_df['depth_primary'].isna()]['tce_depth']
    disc_df['radius(earth)'] = disc_df['pl_rade']
    disc_df.loc[disc_df['radius(earth)'].isna(), 'radius(earth)'] = disc_df[disc_df['radius(earth)'].isna()]['koi_prad']
    disc_df.loc[disc_df['radius(earth)'].isna(), 'radius(earth)'] = disc_df[disc_df['radius(earth)'].isna()]['tce_prad']
    disc_df = disc_df.sort_values(by=["object_id", "period"], ascending=True)
    disc_df.loc[(disc_df['kepoi_name'].isna()) &
                (disc_df['tce_maxmes'] > 7.1), 'type'] = 'tce_secondary'
    disc_df.loc[(disc_df['kepoi_name'].isna()) &
            (disc_df['cent_so_sigma'] > 3), 'type'] = 'tce_centroids_offset'
    disc_df.loc[(disc_df['kepoi_name'].isna()) &
            ((disc_df['source_offset_sigma'] > 3) |
             (disc_df['kic_offset_sigma'] > 3)), 'type'] = 'tce_source_offset'
    disc_df.loc[(disc_df['kepoi_name'].isna()) &
            (disc_df['tce_bin_oedp_stat'] > 9), 'type'] = 'tce_odd_even'
    disc_df.loc[(disc_df['kepoi_name'].isna()) &
                (disc_df['tce_hap_stat'] > disc_df['tce_cap_stat']), 'type'] = 'tce_og'
    disc_df.loc[(disc_df['kepoi_name'].isna()) &
                      (disc_df['type'] == 'tce') &
                      (disc_df['tce_model_snr'] > 3.5), 'type'] = 'tce_candidate'
    #t[t['boot_fap'] > t['boot_mesmean']]
    last_target = ''
    last_period = 0.5
    last_index = -1
    left_match_index = []
    right_match_index = []
    unpaired_index = disc_df.index.values
    for index, target_row in disc_df.iterrows():
        period_diff = 0.2 / (last_period ** (1/2))
        period_diff = 0.05 if period_diff > 0.05 else period_diff
        current_period = target_row['period']
        if last_target == target_row['object_id'] and np.abs(current_period - last_period) < period_diff * last_period:
            right_match_index = right_match_index + [index]
            if last_index not in left_match_index:
                left_match_index = left_match_index + [last_index]
            if last_index in unpaired_index:
                unpaired_index = unpaired_index[unpaired_index != last_index]
            if index in unpaired_index:
                unpaired_index = unpaired_index[unpaired_index != index]
        else:
            last_period = current_period
            last_index = index
        last_target = target_row['object_id']
    #disc_df = disc_df.loc[left_match_index].combine_first(disc_df.loc[right_match_index])
    result_df = disc_df.loc[unpaired_index]
    for i, index in enumerate(left_match_index):
        result = disc_df.loc[index].combine_first(disc_df.loc[right_match_index[i]])
        result['close_period'] = 1
        result_df = pd.concat([result_df, result.to_frame().T], ignore_index=True)
    result_df.drop(result_df[result_df['duration(h)'].isnull()].index, inplace=True)
    result_df.drop(result_df[result_df['object_id'] == 'KIC <NA>'].index, inplace=True)
    # Remove latest KIC from the catalogue
    result_df.drop(result_df[result_df['object_id'] == 'KIC 100001645'].index, inplace=True)
    result_df['ephemeris_contamination'] = 0
    result_df.loc[(result_df['koi_fpflag_ec'] == 1) & (result_df['koi_fpflag_nt'] == 0) &
                  (result_df['koi_fpflag_ss'] == 0) & (result_df['koi_fpflag_co'] == 0), 'ephemeris_contamination'] = 1
    result_df['koi_without_tce'] = 0
    result_df.loc[(~result_df['kepoi_name'].isna()) & (result_df['tce_plnt_num'].isna()), 'koi_without_tce'] = 1
    #result_df.drop(result_df[(np.isnan(result_df['tce_plnt_num']))].index, inplace=True) # discarding rows without DR25 tce
    result_df.to_csv(tces_dir + '/classified_tces.csv')
    results_df_no_validated = result_df.copy()
    results_df_no_validated.loc[results_df_no_validated['disc_refname'].str.contains('Valizadegan', na=False), ['type']] = 'candidate'
    results_df_no_validated.loc[results_df_no_validated['disc_refname'].str.contains('Armstrong et al. 2021', na=False), ['type']] = 'candidate'
    results_df_no_validated.to_csv(tces_dir + '/classified_tces_no_val.csv')
    #disc_df[disc_df['object_id'] == 'KIC 7377200'].iloc[0]['epoch']
    tce_candidate_df = result_df[result_df['type'].isin(['tce_candidate'])]
    tce_candidate_df.to_csv(tces_dir + '/injected_objects_tce_candidates.csv')
    candidate_df = result_df[result_df['type'].isin(['candidate', 'tce_candidate'])]
    candidate_df.to_csv(tces_dir + '/injected_objects_candidates.csv')
    candidate_df_no_val = results_df_no_validated[results_df_no_validated['type'].isin(['candidate', 'tce_candidate'])]
    candidate_df_no_val.to_csv(tces_dir + '/injected_objects_candidates_no_val.csv')
    result_df = result_df[result_df['type'].isin(['fp', 'fa', 'planet', 'planet_transit', 'tce', 'tce_og',
                                                  'tce_secondary', 'tce_source_offset', 'tce_centroids_offset',
                                                  'tce_odd_even'])]
    result_df = result_df[~result_df['radius(earth)'].isna()]
    result_df.to_csv(tces_dir + '/injected_objects_tces.csv')
    results_df_no_validated = results_df_no_validated[results_df_no_validated['type'].isin(['fp', 'fa', 'planet', 'planet_transit', 'tce', 'tce_og',
                                                  'tce_secondary', 'tce_source_offset', 'tce_centroids_offset',
                                                  'tce_odd_even'])]
    results_df_no_validated = results_df_no_validated[~results_df_no_validated['radius(earth)'].isna()]
    results_df_no_validated.to_csv(tces_dir + '/injected_objects_tces_no_val.csv')


def create_targets_multi_df(tces_dir):
    # Merge TCEs DF and cumulative DF ignoring everything not from Kepler or K2
    # Mark TCEs as planet only if they have photometric and rv measurements
    # test_df = pd.read_csv(tces_dir + '/classified_tces.csv', comment='#')
    # test_df = test_df.sort_values(by=["object_id", "period"], ascending=True)
    # test_df = test_df[test_df['type'].isin(['fp', 'planet', 'planet_transit', 'tce'])]
    # test_df.to_csv(tces_dir + '/injected_objects_tces.csv')
    cfps_df = pd.read_csv(tces_dir + '/fpwg_2023.04.25_10.26.55.csv', comment='#')
    kois_df = pd.read_csv(tces_dir + '/cumulative_2023.06.30_13.09.52.csv', comment='#')
    tces_df = pd.read_csv(tces_dir + '/q1_q17_dr25_tce_2023.07.03_23.17.12.csv', comment='#')
    disc_df = pd.read_csv(tces_dir + '/PSCompPars_2023.06.30_07.30.12.csv', comment='#')
    tces_df = tces_df[tces_df['tce_rogue_flag'] == 0]
    kois_df = kois_df[~((kois_df['koi_disposition'] == 'CONFIRMED') & (kois_df['koi_pdisposition'] == 'FALSE POSITIVE'))]
    tces_df['cent_so'] = (tces_df['tce_fwm_srao'] ** 2 + tces_df['tce_fwm_sdeco'] ** 2) ** 0.5
    tces_df['cent_so_err'] = (((tces_df['tce_fwm_srao'] / np.sqrt(tces_df['tce_fwm_srao'] ** 2 + tces_df['tce_fwm_sdeco'] ** 2)) * tces_df['tce_fwm_srao_err']) ** 2 + \
                              ((tces_df['tce_fwm_sdeco'] / np.sqrt(tces_df['tce_fwm_srao'] ** 2 + tces_df['tce_fwm_sdeco'] ** 2)) * tces_df['tce_fwm_sdeco_err']) ** 2) ** 0.5
    tces_df['cent_so_sigma'] = tces_df['cent_so'] / tces_df['cent_so_err']
    tces_df['source_offset_sigma'] = tces_df['tce_dicco_msky'] / tces_df['tce_dicco_msky_err']
    tces_df['kic_offset_sigma'] = tces_df['tce_dikco_msky'] / tces_df['tce_dikco_msky_err']
    # tces_df[~(((tces_df['tce_hap_stat'] / tces_df['tce_cap_stat'] > 1) |
    #            (tces_df['tce_hap_stat'] > 0) & (tces_df['tce_cap_stat'] < 0)) |
    #           (tces_df['tce_maxmes'] > 7.1) |
    #           (tces_df['cent_so_sigma'] > 3) |
    #           (tces_df['source_offset_sigma'] > 3) |
    #           (tces_df['kic_offset_sigma'] > 3)) & (tces_df['tce_model_snr'] > 3.5)]
    kois_df = pd.merge(kois_df, cfps_df, on=['kepoi_name'], how="outer", indicator=True)
    kois_df = kois_df[(kois_df['_merge'] == 'left_only') | (kois_df['_merge'] == 'both')]
    kois_df['kepid'] = kois_df['kepid_x']
    kois_df['pl_name'] = kois_df['kepler_name']
    kois_df['tce_period'] = kois_df['koi_period']
    kois_df['tce_period_round'] = kois_df['tce_period'].round(1)
    tces_df['tce_period_round'] = tces_df['tce_period'].round(1)
    tces_df = pd.merge(tces_df, kois_df, on=['kepid', 'tce_period_round'], how='outer')
    disc_df = disc_df[disc_df['disc_facility'] == 'Kepler']
    disc_df = disc_df[disc_df['discoverymethod'] == 'Transit']
    disc_df = pd.merge(tces_df, disc_df, on='pl_name', how='outer')
    disc_df['multitype'] = ''
    disc_df['close_period'] = 0
    disc_df['type'] = 'tce'
    disc_df.loc[~disc_df['pl_name'].isna(), 'type'] = 'planet'
    disc_df.loc[(disc_df['koi_disposition'] == 'FALSE POSITIVE') & (
        ~disc_df['type'].str.contains('planet')), 'type'] = 'fp_candidate'
    disc_df.loc[(disc_df['koi_disposition'] == 'CONFIRMED'), 'type'] = 'planet_transit'
    disc_df.loc[(disc_df['rv_flag'] > 0) & (disc_df['tran_flag'] > 0) & (
                disc_df['koi_disposition'] == 'CONFIRMED'), 'type'] = 'planet'
    disc_df.loc[(disc_df['rv_flag'] == 0) & (disc_df['tran_flag'] > 0) & (
                disc_df['koi_disposition'] == 'CONFIRMED'), 'type'] = 'planet_transit'
    disc_df.loc[(disc_df['koi_disposition'] == 'FALSE POSITIVE') & (
        ~disc_df['type'].str.contains('planet')), 'type'] = 'fp_candidate'
    disc_df.loc[
        (disc_df['koi_disposition'] == 'CANDIDATE') & (~disc_df['type'].str.contains('planet')), 'type'] = 'candidate'
    disc_df.loc[
        (disc_df['fpwg_disp_status'] == 'CERTIFIED FP') & (~disc_df['type'].str.contains('planet')), 'type'] = 'fp'
    disc_df.loc[
        (disc_df['fpwg_disp_status'] == 'CERTIFIED FA') & (~disc_df['type'].str.contains('planet')), 'type'] = 'fa'
    disc_df['object_id'] = 'KIC ' + disc_df['kepid'].astype('Int64').astype('str')
    disc_df['period'] = disc_df['tce_period_x']
    disc_df.loc[disc_df['period'].isna(), 'period'] = disc_df[disc_df['period'].isna()]['tce_period_y']
    disc_df.loc[disc_df['period'].isna(), 'period'] = disc_df[disc_df['period'].isna()]['pl_orbper']
    disc_df['epoch'] = disc_df['tce_time0bk']
    disc_df.loc[disc_df['epoch'].isna(), 'epoch'] = disc_df[disc_df['epoch'].isna()]['koi_time0bk']
    disc_df.loc[disc_df['epoch'].isna(), 'epoch'] = disc_df[disc_df['epoch'].isna()]['pl_tranmid'] - 2454833.0
    disc_df['duration(h)'] = disc_df['pl_trandur']
    disc_df.loc[disc_df['duration(h)'].isna(), 'duration(h)'] = disc_df[disc_df['duration(h)'].isna()]['koi_duration']
    disc_df.loc[disc_df['duration(h)'].isna(), 'duration(h)'] = disc_df[disc_df['duration(h)'].isna()]['tce_duration']
    disc_df['depth_primary'] = disc_df['pl_trandep'] * 10000
    disc_df.loc[disc_df['depth_primary'].isna(), 'depth_primary'] = disc_df[disc_df['depth_primary'].isna()]['koi_depth']
    disc_df.loc[disc_df['depth_primary'].isna(), 'depth_primary'] = disc_df[disc_df['depth_primary'].isna()]['tce_depth']
    disc_df['radius(earth)'] = disc_df['pl_rade']
    disc_df.loc[disc_df['radius(earth)'].isna(), 'radius(earth)'] = disc_df[disc_df['radius(earth)'].isna()]['koi_prad']
    disc_df.loc[disc_df['radius(earth)'].isna(), 'radius(earth)'] = disc_df[disc_df['radius(earth)'].isna()]['tce_prad']
    disc_df = disc_df.sort_values(by=["object_id", "period"], ascending=True)
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_maxmes'] > 7.1), 'type'] = 'tce_secondary'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_maxmes'] > 7.1), 'multitype'] = \
        disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_maxmes'] > 7.1), 'multitype'] + ',tce_secondary'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['cent_so_sigma'] > 3), 'type'] = 'tce_centroids_offset'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['cent_so_sigma'] > 3), 'multitype'] = \
        disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['cent_so_sigma'] > 3), 'multitype'] + ',tce_centroids_offset'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & ((disc_df['source_offset_sigma'] > 3) | (disc_df['kic_offset_sigma'] > 3)), 'type'] = 'tce_source_offset'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & ((disc_df['source_offset_sigma'] > 3) | (disc_df['kic_offset_sigma'] > 3)), 'multitype'] = \
        disc_df.loc[(disc_df['kepoi_name'].isna()) & ((disc_df['source_offset_sigma'] > 3) | (disc_df['kic_offset_sigma'] > 3)), 'multitype'] + \
        ',tce_source_offset'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_bin_oedp_stat'] > 9), 'type'] = 'tce_odd_even'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_bin_oedp_stat'] > 9), 'multitype'] = \
        disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_bin_oedp_stat'] > 9), 'multitype'] + ',tce_odd_even'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_hap_stat'] > disc_df['tce_cap_stat']), 'type'] = 'tce_og'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_hap_stat'] > disc_df['tce_cap_stat']), 'multitype'] = \
        disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['tce_hap_stat'] > disc_df['tce_cap_stat']), 'multitype'] + ',tce_og'
    disc_df.loc[(disc_df['kepoi_name'].isna()) & (disc_df['type'] == 'tce') & (disc_df['tce_model_snr'] > 3.5), 'type'] = 'tce_candidate'
    disc_df.loc[disc_df['disc_refname'].str.contains('Valizadegan', na=False), ['multitype']] = \
        disc_df.loc[disc_df['disc_refname'].str.contains('Valizadegan', na=False), ['multitype']] + ',planet_validated'
    disc_df.loc[disc_df['disc_refname'].str.contains('Armstrong et al. 2021', na=False), ['multitype']] = \
        disc_df.loc[disc_df['disc_refname'].str.contains('Armstrong et al. 2021', na=False), ['multitype']] + ',planet_validated'
    disc_df.loc[((disc_df['multitype'] == '') | (disc_df['multitype'].isna()) | (disc_df['multitype'] == 'nan')), 'multitype'] = disc_df.loc[((disc_df['multitype'] == '') | (disc_df['multitype'].isna()) | (disc_df['multitype'] == 'nan')), 'type']
    last_target = ''
    last_period = 0.5
    last_index = -1
    left_match_index = []
    right_match_index = []
    unpaired_index = disc_df.index.values
    for index, target_row in disc_df.iterrows():
        period_diff = 0.2 / (last_period ** (1/2))
        period_diff = 0.05 if period_diff > 0.05 else period_diff
        current_period = target_row['period']
        if last_target == target_row['object_id'] and np.abs(current_period - last_period) < period_diff * last_period:
            right_match_index = right_match_index + [index]
            if last_index not in left_match_index:
                left_match_index = left_match_index + [last_index]
            if last_index in unpaired_index:
                unpaired_index = unpaired_index[unpaired_index != last_index]
            if index in unpaired_index:
                unpaired_index = unpaired_index[unpaired_index != index]
        else:
            last_period = current_period
            last_index = index
        last_target = target_row['object_id']
    #disc_df = disc_df.loc[left_match_index].combine_first(disc_df.loc[right_match_index])
    result_df = disc_df.loc[unpaired_index]
    for i, index in enumerate(left_match_index):
        result = disc_df.loc[index].combine_first(disc_df.loc[right_match_index[i]])
        result['close_period'] = 1
        result_df = pd.concat([result_df, result.to_frame().T], ignore_index=True)
    result_df.drop(result_df[result_df['duration(h)'].isnull()].index, inplace=True)
    result_df.drop(result_df[result_df['object_id'] == 'KIC <NA>'].index, inplace=True)
    # Remove latest KIC from the catalogue
    result_df.drop(result_df[result_df['object_id'] == 'KIC 100001645'].index, inplace=True)
    result_df['ephemeris_contamination'] = 0
    result_df.loc[(result_df['koi_fpflag_ec'] == 1) & (result_df['koi_fpflag_nt'] == 0) &
                  (result_df['koi_fpflag_ss'] == 0) & (result_df['koi_fpflag_co'] == 0), 'ephemeris_contamination'] = 1
    result_df['koi_without_tce'] = 0
    result_df.loc[(~result_df['kepoi_name'].isna()) & (result_df['tce_plnt_num'].isna()), 'koi_without_tce'] = 1
    #result_df.drop(result_df[(np.isnan(result_df['tce_plnt_num']))].index, inplace=True) # discarding rows without DR25 tce
    result_df.to_csv(tces_dir + '/classified_tces_multi.csv')
    #disc_df[disc_df['object_id'] == 'KIC 7377200'].iloc[0]['epoch']
    result_df = result_df[result_df['type'].isin(['fp', 'fa', 'planet', 'planet_transit', 'tce', 'tce_og',
                                                  'tce_secondary', 'tce_source_offset', 'tce_centroids_offset',
                                                  'tce_odd_even'])]
    result_df = result_df[~result_df['radius(earth)'].isna()]
    #TODO this next statement is to prevent confirmed planets not matching any tce nor koi
    #result_df = result_df[~result_df['object_id'].str.contains('<NA')]
    result_df.to_csv(tces_dir + '/injected_objects_tces_multi.csv')


def create_synth_targets_df(tces_dir):
    inj1_df = pd.read_csv(tces_dir + '/kplr_dr25_inj1_tces.csv', comment='#')
    inj1_df['type'] = 'inj1'
    inj2_df = pd.read_csv(tces_dir + '/kplr_dr25_inj2_tces.csv', comment='#')
    inj2_df['type'] = 'inj2'
    inj3_df = pd.read_csv(tces_dir + '/kplr_dr25_inj3_tces.csv', comment='#')
    inj3_df['type'] = 'inj3'
    inv_df = pd.read_csv(tces_dir + '/kplr_dr25_inv_tces.csv', comment='#')
    inv_df['type'] = 'inv'
    scr1_df = pd.read_csv(tces_dir + '/kplr_dr25_scr1_tces.csv', comment='#')
    scr1_df['type'] = 'scr1'
    scr2_df = pd.read_csv(tces_dir + '/kplr_dr25_scr2_tces.csv', comment='#')
    scr2_df['type'] = 'scr2'
    scr3_df = pd.read_csv(tces_dir + '/kplr_dr25_scr3_tces.csv', comment='#')
    scr3_df['type'] = 'scr3'
    synth_df = pd.concat([inj1_df, inj2_df, inj3_df, inv_df, scr1_df, scr2_df, scr3_df], ignore_index=True)
    synth_df['duration(h)'] = synth_df['duration']
    synth_df['depth_primary'] = synth_df['depth']
    synth_df['radius(earth)'] = synth_df['Rp']
    synth_df.rename(columns={'KIC': 'kepid'}, inplace=True)
    synth_df['object_id'] = 'KIC ' + synth_df.loc[:, 'kepid'].astype(str)
    synth_df.to_csv(tces_dir + '/classified_tces_synthetic.csv')

def create_tess_cps_df(df_dir):
    tois_df = pd.read_csv('/home/martin/.sherlockpipe/tois.csv')
    tois_df = tois_df.loc[(tois_df["Disposition"] == "KP") | (tois_df["Disposition"] == "CP")]
    tois_df['object_id'] = tois_df['Object Id']
    tois_df['kepid'] = tois_df['Object Id'].str[4:]
    tois_df['period'] = tois_df['Period (days)']
    tois_df['epoch'] = tois_df['Epoch (BJD)']
    tois_df['duration(h)'] = tois_df['Duration (hours)']
    tois_df['depth_primary'] = tois_df['Depth (ppm)']
    tois_df['radius(earth)'] = tois_df['Planet Radius (R_Earth)']
    tois_df = tois_df.loc[(~tois_df['period'].isna()) & (tois_df['period'] > 0.0)]
    tois_df.to_csv(f"{df_dir}/tess_cps.csv", index=False)

def create_tess_fps_df(df_dir):
    tois_df = pd.read_csv('/home/martin/.sherlockpipe/tois.csv')
    tois_df = tois_df.loc[(tois_df["Disposition"] == "FP") | (tois_df["Disposition"] == "FA")]
    tois_df['object_id'] = tois_df['Object Id']
    tois_df['kepid'] = tois_df['Object Id'].str[4:]
    tois_df['period'] = tois_df['Period (days)']
    tois_df['epoch'] = tois_df['Epoch (BJD)']
    tois_df['duration(h)'] = tois_df['Duration (hours)']
    tois_df['depth_primary'] = tois_df['Depth (ppm)']
    tois_df['radius(earth)'] = tois_df['Planet Radius (R_Earth)']
    tois_df = tois_df.loc[(~tois_df['period'].isna()) & (tois_df['period'] > 0.0)]
    tois_df.to_csv(f"{df_dir}/tess_cps.csv", index=False)

def create_tess_tces_df(df_dir):
    df = None
    for file in sorted(glob.glob(f'{df_dir}tess*tcestats.csv')):
        if df is None:
            df = pd.read_csv(file, comment='#')
        else:
            df = pd.concat([df, pd.read_csv(file, comment='#')], ignore_index=True)
    tois_df = pd.read_csv('/home/martin/.sherlockpipe/tois.csv')
    removed_tces = 0
    df['type'] = 'tce'
    for index, row in df.copy().iterrows():
        if len(tois_df.loc[(np.abs(tois_df['Period (days)'] - row['tce_period']) < 0.05) & (tois_df['Object Id'].str[4:].astype(int) == row['ticid'])]) > 0:
            df.loc[df['tceid'] == row['tceid'], 'type'] = 'toi'
            removed_tces = removed_tces + 1
    print("Removed tces " + str(removed_tces))
    df['object_id'] = "TIC " + df['ticid'].apply(lambda x: str(x))
    df['kepid'] = df['ticid']
    df['period'] = df['tce_period']
    df['epoch'] = df['tce_time0bt']
    df['duration(h)'] = df['tce_duration']
    df['depth_primary'] = df['tce_depth']
    df['radius(earth)'] = df['tce_prad']
    df.to_csv(df_dir + '/tces.csv')

def download_synth_targets(download_dir, wget_files, df_file):
    df = pd.read_csv(df_file, comment='#')
    os.chdir(download_dir)
    pattern = r'kplr(\d+).*'
    pattern_date = r'(\d+)\_INJECTED.*'
    for wget_file in wget_files:
        with open(wget_file, 'r') as file:
            for line in file:
                match = re.search(pattern, line)
                if match is not None:
                    kepid = int(match.group(1))
                    rows = df[df['KIC'] == kepid]
                    for index, row in rows.iterrows():
                        if 'inj' in row['type']:
                            match_date = re.search(pattern_date, line)
                            lc_file = f'{download_dir}/kplr{kepid:09}-{match_date.group(1)}_INJECTED-' + row['type'] + '_llc.fits'
                            if not os.path.exists(lc_file):
                                print(f"Downloading: {line}")
                                os.system(line)
                                compressed_file = f'{lc_file}.gz'
                                if os.path.exists(compressed_file):
                                    with gzip.open(compressed_file, 'rb') as f_in:
                                        with open(lc_file, 'wb') as f_out:
                                            shutil.copyfileobj(f_in, f_out)
                                            os.remove(compressed_file)


def create_target_csvs(csv, target_dir, cache_dir, cores=multiprocessing.cpu_count() - 1, force=False, ids=None,
                       mode='all', mission_prefix='KIC'):
    tces_df = pd.read_csv(csv, comment='#')
    tces_df = tces_df.sort_values(by=["object_id"], ascending=True)
    #tces_df = tces_df.drop_duplicates(subset=["kepid"], keep='last')
    tces_df = tces_df.sort_values(by=["period"], ascending=True)
    #tces_df = tces_df.sample(frac=1).reset_index(drop=True)
    inputs = []
    if ids is not None:
        tces_df = tces_df[tces_df['object_id'].isin(ids)]
    tces_df = tces_df.reset_index(drop=True)
    ranges = range(0, len(tces_df))
    print("Total number of targets is " + str(len(tces_df)))
    for index in ranges:
        tce_row = tces_df.iloc[index]
        inputs = inputs + [CreateTargetInput(mission_prefix, tce_row['object_id'], None, target_dir, None, tce_row, cache_dir, index, force=force, mode=mode)]
    with Pool(processes=cores, maxtasksperchild=1) as pool:
        pool.map(create_target_csv, inputs, chunksize=1)


def create_synth_csvs(csv, lcs_dir, target_dir, cache_dir, cores=multiprocessing.cpu_count() - 1, force=False, ids=None, mode='all'):
    synth_df = pd.read_csv(csv, comment='#')
    synth_df = synth_df.sort_values(by=["kepid"], ascending=True)
    inputs = []
    if ids is not None:
        synth_df = synth_df[synth_df['kepid'].isin(int(id) for id in ids)]
    synth_df = synth_df.reset_index(drop=True)
    ranges = range(0, len(synth_df))
    print("Total number of targets is " + str(len(synth_df)))
    for index in ranges:
        tce_row = synth_df.iloc[index]
        kepid = int(tce_row['kepid'])
        matching_files = None
        if 'inj' in tce_row['type']:
            path_pattern = f"{lcs_dir}/kplr{kepid:09}-*.fits"
            matching_files = sorted(glob.glob(path_pattern))
        inputs = inputs + [CreateTargetInput('KIC', 'KIC ' + str(kepid), matching_files, target_dir, None, tce_row, cache_dir, index, force=force, mode=mode)]
    with Pool(processes=cores, maxtasksperchild=1) as pool:
        pool.map(create_target_csv, inputs, chunksize=1)

def create_triceratops_probs(csv, target_dir, cache_dir, cores=multiprocessing.cpu_count() - 1, force=False, ids=None):
    tces_df = pd.read_csv(csv, comment='#')
    tces_df = tces_df.sort_values(by=["kepid"], ascending=True)
    # tces_df = tces_df.drop_duplicates(subset=["kepid"], keep='last')
    #tces_df = tces_df.sample(frac=1).reset_index(drop=True)
    if not force and 'tric_tp_total' in tces_df.columns:
        tces_df = tces_df.loc[tces_df['tric_tp_total'].isna()]
    print("Total number of targets is " + str(len(tces_df)))
    inputs = []
    if ids is not None:
        tces_df = tces_df[tces_df['object_id'].isin(ids)]
    for index, tce_row in tces_df.iterrows():
        inputs = inputs + [
            CreateTargetInput('KIC', tce_row['object_id'], None, target_dir, None, tce_row, cache_dir, force=force)]
    with Pool(processes=cores, maxtasksperchild=1) as pool:
        pool.map(create_triceratops_prob, inputs, chunksize=1)



import logging

import tensorflow as tf
from keras.layers import SpatialDropout1D
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.layers import Dense, Dropout
from tensorflow.keras.regularizers import l1_l2


class DynamicRegularizationAndLearningRateCallback(Callback):
    def __init__(self, model, l1=0.1, l2=0.1, dropout=0.1, spatial_dropout=0.1, l1_adjustment_factor=0.1,
                 l2_adjustment_factor=0.1, dropout_adjustment_factor=0.1, patience=3, lr_adjustment_factor=0.5):
        super(DynamicRegularizationAndLearningRateCallback, self).__init__()
        self.model = model
        self.l1 = l1
        self.l2 = l2
        self.dropout = dropout
        self.spatial_dropout = spatial_dropout
        self.l1_adjustment_factor = l1_adjustment_factor
        self.l2_adjustment_factor = l2_adjustment_factor
        self.dropout_adjustment_factor = dropout_adjustment_factor
        self.lr_adjustment_factor = lr_adjustment_factor
        self.patience = patience
        self.wait_val = 0
        self.wait_train = 0
        self.best_val_loss = float('inf')
        self.best_train_loss = float('inf')

    def on_epoch_end(self, epoch, logs=None):
        current_val_loss = logs.get('val_loss')
        current_train_loss = logs.get('loss')
        current_lr = tf.keras.backend.get_value(self.model.optimizer.lr)
        if current_val_loss < self.best_val_loss:
            self.best_val_loss = current_val_loss
            self.wait_val = 0
        else:
            self.wait_val += 1
        if current_train_loss < self.best_train_loss:
            self.best_train_loss = current_train_loss
            self.wait_train = 0
        else:
            self.wait_train += 1
        if self.wait_val >= self.patience:
            if current_val_loss > current_train_loss:
                self.l1 = self.l1 * (1 + self.l1_adjustment_factor)
                self.l2 = self.l2 * (1 + self.l2_adjustment_factor)
                self.dropout = self.dropout * (1 + self.dropout_adjustment_factor)
                self.spatial_dropout = self.spatial_dropout * (1 + self.dropout_adjustment_factor)
            logging.info(f"\nAdjusting regularization: L1={self.l1}, L2={self.l2}, Dropout={self.dropout}, "
                         f"SpatialDropout={self.spatial_dropout}")
            for layer in self.model.layers:
                if isinstance(layer, Dropout):
                    layer.rate = self.dropout
                if isinstance(layer, SpatialDropout1D):
                    layer.rate = self.dropout
                if isinstance(layer, Dense):
                    layer.kernel_regularizer = l1_l2(l1=self.l1, l2=self.l2)
            self.wait_val = 0
        if self.wait_train >= self.patience:
            new_lr = current_lr * self.lr_adjustment_factor
            tf.keras.backend.set_value(self.model.optimizer.lr, new_lr)
            logging.info(f"\nAdjusting learning rate to: {new_lr}")
            self.wait_train = 0
        if self.wait_val >= self.patience or self.wait_train >= self.patience:
            self.model.compile(optimizer=self.model.optimizer, loss=self.model.loss, metrics=self.model.metrics)


import tensorflow as tf
from keras_core.src.layers import LeakyReLU
#from keras.layers import MultiHeadAttention, LeakyReLU
#from keras.src.layers import Dense
from keras_nlp.src.layers import TransformerEncoder

from exoml.ml.layers.dropout import AdaptiveStdDropout


@tf.keras.utils.register_keras_serializable()
class ClassToken(tf.keras.layers.Layer):
    """Append a class token to an input layer."""

    def build(self, input_shape):
        cls_init = tf.zeros_initializer()
        self.hidden_size = input_shape[-1]
        self.cls = tf.Variable(
            name="cls",
            initial_value=cls_init(shape=(1, 1, self.hidden_size), dtype="float32"),
            trainable=True,
        )

    def call(self, inputs):
        batch_size = tf.shape(inputs)[0]
        cls_broadcasted = tf.cast(
            tf.broadcast_to(self.cls, [batch_size, 1, self.hidden_size]),
            dtype=inputs.dtype,
        )
        return tf.concat([cls_broadcasted, inputs], 1)

    def get_config(self):
        config = super().get_config()
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)


@tf.keras.utils.register_keras_serializable()
class AddPositionEmbs(tf.keras.layers.Layer):
    """Adds (optionally learned) positional embeddings to the inputs."""

    def build(self, input_shape):
        assert (
            len(input_shape) == 3
        ), f"Number of dimensions should be 3, got {len(input_shape)}"
        self.pe = tf.Variable(
            name="pos_embedding",
            initial_value=tf.random_normal_initializer(stddev=0.06)(
                shape=(1, input_shape[1], input_shape[2])
            ),
            dtype="float32",
            trainable=True,
        )

    def call(self, inputs):
        return inputs + tf.cast(self.pe, dtype=inputs.dtype)

    def get_config(self):
        config = super().get_config()
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)


# @tf.keras.utils.register_keras_serializable()
# class TransformerEncoder(tf.keras.layers.Layer):
#     """Implements a Transformer block."""
#
#     def __init__(self, *args, num_heads, mlp_dim, hyperparams, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.num_heads = num_heads
#         self.mlp_dim = mlp_dim
#         self.hyperparams = hyperparams
#         self.regularizer_value = 0.1
#
#     def build(self, input_shape):
#         self.att = MultiHeadAttention(num_heads=self.num_heads,
#             key_dim=input_shape[-1],
#             name="MultiHeadDotProductAttention_1"
#         )
#         self.mlpblock = tf.keras.Sequential(
#             [
#                 tf.keras.layers.Dense(self.mlp_dim, activation=LeakyReLU(), name=f"{self.name}/Dense_0",
#                                       activity_regularizer=tf.keras.regularizers.L1L2(
#                                           l1=self.hyperparams.l1_regularization,
#                                           l2=self.hyperparams.l2_regularization)
#                                       ),
#                 AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
#                                    max_rate=self.hyperparams.dropout_max_rate),
#                 tf.keras.layers.Dense(input_shape[-1], name=f"{self.name}/Dense_1", activation=LeakyReLU(),
#                                       activity_regularizer=tf.keras.regularizers.L1L2(
#                                           l1=self.hyperparams.l1_regularization,
#                                           l2=self.hyperparams.l2_regularization)
#                                       )
#             ],
#             name="MlpBlock_3",
#         )
#         self.layernorm1 = tf.keras.layers.LayerNormalization(
#             epsilon=1e-6, name="LayerNorm_1"
#         )
#         self.layernorm2 = tf.keras.layers.LayerNormalization(
#             epsilon=1e-6, name="LayerNorm_2"
#         )
#         self.layernorm3 = tf.keras.layers.LayerNormalization(
#             epsilon=1e-6, name="LayerNorm_3"
#         )
#         self.dropout_layer = AdaptiveStdDropout(rate=self.hyperparams.dropout_rate,
#                                          max_rate=self.hyperparams.dropout_max_rate)
#
#     def call(self, inputs, training=True, return_attention_scores=True):
#         x = inputs
#         #x = self.layernorm1(x)
#         x = self.att(x, x, training=training, return_attention_scores=return_attention_scores)
#         x = self.dropout_layer(x, training=training)
#         x = x + inputs
#         y = self.layernorm2(x)
#         y = self.mlpblock(y, training=training)
#         return self.layernorm3(x + y)
#
#     def get_config(self):
#         config = super().get_config()
#         config.update(
#             {
#                 "num_heads": self.num_heads,
#                 "mlp_dim": self.mlp_dim,
#                 "dropout": self.hyperparams.dropout_rate,
#             }
#         )
#         return config
#
#     @classmethod
#     def from_config(cls, config):
#         return cls(**config)


@tf.keras.utils.register_keras_serializable()
class Transformer(tf.keras.layers.Layer):
    def __init__(self, *args, num_heads, mlp_dim, hyperparams, num_blocks, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_heads = num_heads
        self.mlp_dim = mlp_dim
        self.hyperparams = hyperparams
        self.num_blocks = num_blocks
        self.transformer_encoders = []
        self.regularizer_value = 0.1

    def build(self, input_shape):
        for n in range(self.num_blocks):
            self.transformer_encoders = self.transformer_encoders + \
                                         [TransformerEncoder(
                self.mlp_dim,
                self.num_heads,
                dropout=self.hyperparams.dropout_rate,
                activation=LeakyReLU(),
                layer_norm_epsilon=self.hyperparams.transformer_layer_norm_epsilon,
                kernel_initializer="glorot_uniform",
                bias_initializer="zeros",
                normalize_first=False
            )]
            # self.transformer_encoders = self.transformer_encoders + [TransformerEncoder(
            #     num_heads=self.num_heads,
            #     mlp_dim=self.mlp_dim,
            #     hyperparams=self.hyperparams,
            #     name=f"Transformer/encoderblock_{n}"
            # )]

    def call(self, inputs, training=True):
        result = inputs
        for n in range(self.num_blocks):
            result = self.transformer_encoders[n](result)
        return result

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_heads": self.num_heads,
                "mlp_dim": self.mlp_dim,
                "dropout": self.hyperparams.dropout_rate,
                "num_blocks": self.num_blocks,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)


@tf.keras.utils.register_keras_serializable()
class TransformerClassifier(tf.keras.layers.Layer):
    def __init__(self, *args, transformer_input_size, patch_size, num_heads, mlp_dim, hyperparams, num_blocks, classes,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.transformer_input_size = transformer_input_size
        self.patch_size = patch_size
        self.num_heads = num_heads
        self.mlp_dim = mlp_dim
        self.hyperparams = hyperparams
        self.num_blocks = num_blocks
        self.classes = classes
        self.regularizer_value = 0.1

    def build(self, input_shape):
        # The linear projection could be a conv1d if the input was an image to be patched in pieces. As it is not,
        # we don't need a convolutional layer because for us a patch is of kernel=1
        # self.linear_proj = tf.keras.layers.Conv1D(
        #     filters=self.transformer_input_size,
        #     kernel_size=self.patch_size,
        #     strides=self.patch_size,
        #     padding="same",
        #     name="embedding"
        # )
        #self.linear_proj = tf.keras.layers.Dense(self.transformer_input_size)
        # self.linear_proj = tf.keras.layers.Dense(self.transformer_input_size, activation="relu")
        #self.class_token = ClassToken(name="class_token")
        #self.add_pos_embedding = AddPositionEmbs(name="add_pos_embedding")
        self.transformer = Transformer(num_heads=self.num_heads, mlp_dim=self.mlp_dim,
                                       hyperparams=self.hyperparams, num_blocks=self.num_blocks)
        self.layer_norm = tf.keras.layers.LayerNormalization(epsilon=self.hyperparams.transformer_layer_norm_epsilon, name="Transformer/encoder_norm")
        activation = "softmax" if self.classes > 1 else "sigmoid"
        self.output_flatten = tf.keras.layers.Flatten()
        self.head = tf.keras.layers.Dense(self.classes, name="head", activation=activation,
                                          activity_regularizer=tf.keras.regularizers.L1L2(
                                              l1=self.hyperparams.l1_regularization,
                                              l2=self.hyperparams.l2_regularization)
                                          )


    def call(self, inputs):
        # result = self.linear_proj(inputs)
        # Assuming we don't need to reshape our data as we don't need to flatten the resulting image of the linear_proj
        #result = tf.keras.layers.Reshape((result.shape[1] * result.shape[2], self.transformer_input_size))(result)
        #result = self.class_token(result)
        #result = self.add_pos_embedding(result)
        result = self.transformer(inputs)
        # result = tf.keras.layers.Lambda(
        #      lambda v: v[:, 0],
        #      name="ExtractToken")(result)
        result = self.layer_norm(result)
        result = self.output_flatten(result)
        result = self.head(result)
        #result = self.dropout_layer(result, training=training)
        return result



import os
import shutil
import unittest


class TestsExoml(unittest.TestCase):
    def test_setup_files(self):
        self.assertTrue(True)
        #IATSON_planet()


if __name__ == '__main__':
    unittest.main()



import numbers

import tensorflow as tf
import keras
from keras import backend
from keras.layers import Dropout


def smart_cond(pred, true_fn=None, false_fn=None, name=None):
    """Took function from abandonned tensorflow-addons

    Return either `true_fn()` if predicate `pred` is true else `false_fn()`.

    If `pred` is a bool or has a constant value, we return either `true_fn()`
    or `false_fn()`, otherwise we use `tf.cond` to dynamically route to both.

    Args:
      pred: A scalar determining whether to return the result of `true_fn` or
        `false_fn`.
      true_fn: The callable to be performed if pred is true.
      false_fn: The callable to be performed if pred is false.
      name: Optional name prefix when using `tf.cond`.

    Returns:
      Tensors returned by the call to either `true_fn` or `false_fn`.

    Raises:
      TypeError: If `true_fn` or `false_fn` is not callable.
    """
    if isinstance(pred, tf.Variable):
        return tf.cond(pred, true_fn=true_fn, false_fn=false_fn, name=name)
    return tf.__internal__.smart_cond.smart_cond(
        pred, true_fn=true_fn, false_fn=false_fn, name=name
    )


class AdaptiveStdDropout(Dropout):
    """
    Experimental dropout increasing the base rate by the inverse of the tanh of the standard deviation of the input
    layer.
    """
    def __init__(self, rate, max_rate=0.8, noise_shape=None, seed=None, **kwargs):
        super().__init__(rate, noise_shape, seed, **kwargs)
        assert rate <= max_rate < 1
        self.max_rate = max_rate

    def call(self, inputs, training=None):
        if isinstance(self.rate, numbers.Real) and self.rate == 0:
            return tf.identity(inputs)

        if training is None:
            training = backend.learning_phase()
        dropout_rate = self.rate
        if self.max_rate > self.rate:
            dropout_rate = tf.math.reduce_std(inputs)
            dropout_rate = tf.math.subtract(1.0, tf.math.tanh(dropout_rate))
            dropout_rate = tf.add(tf.math.multiply(dropout_rate, self.max_rate - self.rate), self.rate)
            max_rate_tensor = tf.convert_to_tensor(self.max_rate, dtype=tf.float32)
            dropout_rate = tf.cond(tf.greater(dropout_rate, self.max_rate), lambda: max_rate_tensor, lambda: dropout_rate)

        def dropped_inputs():
            return self._random_generator.dropout(
                inputs, dropout_rate, noise_shape=self._get_noise_shape(inputs)
            )

        output = smart_cond(
            training, dropped_inputs, lambda: tf.identity(inputs)
        )
        return output

    def get_config(self):
        config = {
            "rate": self.rate,
            "noise_shape": self.noise_shape,
            "seed": self.seed,
            "max_rate": self.max_rate
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))




import multiprocessing

import numpy as np
import pandas as pd
import keras
from keras.layers import LeakyReLU
from sklearn.utils import shuffle

from exoml.ml.model.base_model import CategoricalPredictionSetStats
from exoml.ml.model.imbalanced_binary_model import ImbalancedBinaryModel
from transpot.transpot_generator import TranspotModelGenerator

class TRANSPOT(ImbalancedBinaryModel):
    def __init__(self, class_ids, input_size=(20610, 13)) -> None:
        super().__init__('TRANSPOT', input_size, class_ids)

    def _get_flux_conv_branch(self):
        # (time, flux, [detrended_flux1 ... detrended_flux5], flux_model, centroidx, centroidy, motionx, motiony, bck)
        # flux model by transit params and stellar params
        autoencoder_layer1_strides = 10
        autoencoder_layer1_filters = 5000
        autoencoder_layer1_ks = 100
        autoencoder_layer2_strides = 5
        autoencoder_layer2_filters = 1250
        autoencoder_layer2_ks = 33
        autoencoder_layer3_strides = 5
        autoencoder_layer3_filters = 420
        autoencoder_layer3_ks = 15
        autoencoder_layer4_strides = 2
        autoencoder_layer4_filters = 128
        autoencoder_layer4_ks = 9
        autoencoder_layer5_strides = 2
        autoencoder_layer5_filters = 64
        autoencoder_layer5_ks = 7
        autoencoder_layer6_strides = 1
        autoencoder_layer6_filters = 32
        autoencoder_layer6_ks = 5
        flux_input = keras.Input(shape=(self.input_size))
        flux_branch = keras.layers.SpatialDropout1D(rate=0.1)(flux_input)
        flux_branch = keras.layers.Conv1D(filters=autoencoder_layer1_filters, kernel_size=autoencoder_layer1_ks,
                                                padding="same", activation=LeakyReLU(0.01))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=50, strides=autoencoder_layer1_strides, padding="same")(
            flux_branch)
        flux_branch = keras.layers.Dropout(rate=0.1)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=autoencoder_layer2_filters, kernel_size=autoencoder_layer2_ks,
                                                padding="same", activation=LeakyReLU(0.01))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=20, strides=autoencoder_layer2_strides, padding="same")(
            flux_branch)
        flux_branch = keras.layers.Dropout(rate=0.1)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=autoencoder_layer3_filters, kernel_size=autoencoder_layer3_ks,
                                                padding="same", activation=LeakyReLU(0.01))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=15, strides=autoencoder_layer3_strides, padding="same")(
            flux_branch)
        flux_branch = keras.layers.Dropout(rate=0.1)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=autoencoder_layer4_filters, kernel_size=autoencoder_layer4_ks,
                                                padding="same", activation=LeakyReLU(0.01))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=10, strides=autoencoder_layer4_strides, padding="same")(
            flux_branch)
        flux_branch = keras.layers.Dropout(rate=0.1)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=autoencoder_layer5_filters, kernel_size=autoencoder_layer5_ks,
                                                padding="same", activation=LeakyReLU(0.01))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=4, strides=autoencoder_layer5_strides, padding="same")(
            flux_branch)
        flux_branch = keras.layers.Dropout(rate=0.1)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=autoencoder_layer6_filters, kernel_size=autoencoder_layer6_ks,
                                                padding="same", activation=LeakyReLU(0.01))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=3, strides=autoencoder_layer6_strides, padding="same")(
            flux_branch)
        flux_branch = keras.layers.Dropout(rate=0.1)(flux_branch)
        #flux_branch = keras.layers.LayerNormalization()(flux_branch)
        return flux_input, flux_branch

    def _get_flux_model_branch(self):
        leaky_relu_alpha = 0.01
        flux_input, flux_branch = self._get_flux_conv_branch()
        flux_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = keras.layers.Dropout(rate=0.1)(flux_branch)
        flux_branch = keras.layers.Dense(16, activation='linear')(flux_branch)
        #flux_branch = keras.layers.LayerNormalization()(flux_branch)
        flux_branch = keras.Model(inputs=[flux_input], outputs=flux_branch)
        return flux_branch

    def build(self):
        leaky_relu_alpha = 0.01
        stellar_model_input = keras.Input(shape=(11, 1), name="stellar_model")
        stellar_model_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha), name="stellar-first")(stellar_model_input)
        stellar_model_branch = keras.layers.Dropout(rate=0.1, name="stellar-first-dropout-0.1")(stellar_model_branch)
        stellar_model_branch = keras.layers.Dense(16, activation="relu", name="stellar-refinement")(
            stellar_model_branch)
        flux_model_branch = self._get_flux_model_branch()
        final_branch = keras.layers.Concatenate(axis=1)([stellar_model_branch, flux_model_branch.output])
        final_branch = keras.layers.GlobalAveragePooling1D()(final_branch)
        final_branch = keras.layers.Dense(64, activation=LeakyReLU(leaky_relu_alpha), name="final-dense1")(final_branch)
        final_branch = keras.layers.Dropout(rate=0.1)(final_branch)
        final_branch = keras.layers.Dense(32, activation=LeakyReLU(leaky_relu_alpha), name="final-dense2")(final_branch)
        final_branch = keras.layers.Dropout(rate=0.1)(final_branch)
        final_branch = keras.layers.Dense(1, activation="sigmoid", name="final-dense-softmax")(final_branch)
        inputs = [stellar_model_input] + flux_model_branch.inputs
        self.set_model(keras.Model(inputs=inputs, outputs=final_branch, name="TRANSPOT"))
        return self

    def load_training_set(self, **kwargs):
        training_dir = kwargs.get('training_dir')
        injected_objects_df = pd.read_csv(training_dir + "/injected_objects.csv", index_col=None)
        injected_objects_df = injected_objects_df.sort_values(by=['TIC ID', 'type'])
        #TODO remove 'none' types when there is more than one entry for same TIC (keeping first as none is the latest
        #  from bckEB, EB, planet and none
        injected_objects_df['type'] = injected_objects_df['type'].astype('category')
        injected_objects_df['type'] = injected_objects_df['type'].cat.set_categories(['bckEB', 'EB', 'planet', 'none'],
                                                                                     ordered=True)
        injected_objects_df.sort_values(by=['TIC ID', 'type'], inplace=True)
        injected_objects_df = injected_objects_df.drop_duplicates(subset='TIC ID', keep="first")
        injected_objects_df = shuffle(injected_objects_df)
        injected_objects_df.reset_index(inplace=True, drop=True)
        return injected_objects_df

    def instance_generator(self, dataset, dir, batch_size, input_sizes, zero_epsilon):
        return TranspotModelGenerator(dataset, dir, batch_size, input_sizes, zero_epsilon)

    def predict(self, input):
        prediction = self.model(input, training=False)
        return self.predict_threshold([prediction])

    def predict_batch(self, inputs, expected_outputs=None, batch_size=20, cores=multiprocessing.cpu_count() - 1):
        predictions = self.model.predict(inputs, use_multiprocessing=True, batch_size=batch_size, workers=cores)
        max_prediction_indexes, max_prediction_values = self.predict_threshold(predictions)
        if expected_outputs is not None:
            bckeb_stats = self.test_metrics(expected_outputs, predictions, 0)
            eb_stats = self.test_metrics(expected_outputs, predictions, 1)
            planet_stats = self.test_metrics(expected_outputs, predictions, 2)
            none_stats = self.test_metrics(expected_outputs, predictions, 3)
            return max_prediction_indexes, max_prediction_values, [bckeb_stats, eb_stats, planet_stats, none_stats]
        else:
            return max_prediction_indexes, max_prediction_values

    def predict_threshold(self, predictions, threshold=[0.5, 0.5, 0.5, 0.5]):
        max_prediction_indexes = [np.argmax(prediction) for prediction in predictions]
        max_prediction_values = [predictions[index][max_prediction_index]
                                 if predictions[index][max_prediction_index] > threshold[predictions[index][max_prediction_index]]
                                 else -1
                                 for index, max_prediction_index in enumerate(max_prediction_indexes)]
        max_prediction_indexes[max_prediction_values == -1] = -1
        return max_prediction_indexes, max_prediction_values

    def predict_test_set(self, training_dir, test_set_file, batch_size=20, zero_epsilon=1e-7):
        test_dataset = pd.read_csv(test_set_file)
        lcs_dir = training_dir + '/lcs/'
        testing_batch_generator = TranspotModelGenerator(test_dataset, lcs_dir, batch_size, self.input_size,
                                                       zero_epsilon)
        expected_outputs = None
        for i in np.arange(0, len(test_dataset) // batch_size + 1, 1):
            input, expected_output = testing_batch_generator.__getitem__(i)
            expected_outputs = np.concatenate((expected_outputs, expected_output)) if expected_outputs is not None else expected_output
        return self.predict_batch(testing_batch_generator, expected_outputs, batch_size)

    def test_metrics(self, expected_outputs, max_prediction_indexes, label_index):
        set_count = len(expected_outputs)
        max_output_indexes = [np.max(expected_output) for expected_output in expected_outputs]
        matches_indexes = np.argwhere(max_output_indexes == max_prediction_indexes).flatten()
        matches_count = len(matches_indexes)
        accuracy = matches_count / set_count
        expected_label_indexes = np.argwhere(max_output_indexes == label_index).flatten()
        output_label_indexes = np.argwhere(max_output_indexes == label_index).flatten()
        label_tp_indexes = np.intersect1d(expected_label_indexes, output_label_indexes)
        label_fp_indexes = output_label_indexes[~np.isin(output_label_indexes, expected_label_indexes)]
        label_fn_indexes = expected_label_indexes[~np.isin(expected_label_indexes, output_label_indexes)]
        label_tp = len(label_tp_indexes)
        label_fp = len(label_fp_indexes)
        label_fn = len(label_fn_indexes)
        precision = label_tp / (label_tp + label_fp)
        recall = label_tp / (label_tp + label_fn)
        return CategoricalPredictionSetStats(label_tp, label_fp, label_fn, accuracy, precision, recall)



import tensorflow as tf


class BranchDropoutLayer(tf.keras.layers.Layer):
    def __init__(self, dropout_rate=0.5, **kwargs):
        super(BranchDropoutLayer, self).__init__(**kwargs)
        self.dropout_rate = dropout_rate  # Probability to drop a branch

    def call(self, inputs, training=None):
        if training:
            mask = tf.random.uniform(shape=(), minval=0, maxval=1) > self.dropout_rate
            mask = tf.cast(mask, dtype=inputs.dtype)  # Convert mask to same dtype as inputs
            # Multiply the input by the mask (0 drops the input, 1 keeps it)
            return inputs * mask
        else:
            return inputs

    def get_config(self):
        config = super(BranchDropoutLayer, self).get_config()
        config.update({"dropout_rate": self.dropout_rate})
        return config